from datetime import date
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional
from zlib import crc32

import deprecation
import numpy as np
import orjson as json
from upolygon import draw_polygon, rle_encode

import darwin.datatypes as dt
from darwin.exporter.formats.numpy_encoder import NumpyEncoder
from darwin.utils import convert_polygons_to_sequences
from darwin.version import __version__

DEPRECATION_MESSAGE = """

This function is going to be turned into private. This means that breaking 
changes in its interface and implementation are to be expected. We encourage using ``export`` 
instead of calling this low-level function directly.

"""


def export(annotation_files: Iterator[dt.AnnotationFile], output_dir: Path) -> None:
    """
    Exports the given ``AnnotationFile``\\s into the coco format inside of the given ``output_dir``.

    Parameters
    ----------
    annotation_files : Iterator[dt.AnnotationFile]
        The ``AnnotationFile``\\s to be exported.
    output_dir : Path
        The folder where the new coco file will be.
    """
    output = _build_json(list(annotation_files))
    output_file_path = (output_dir / "output").with_suffix(".json")
    with open(output_file_path, "w") as f:
        op = json.dumps(output, option=json.OPT_INDENT_2 | json.OPT_SERIALIZE_NUMPY).decode("utf-8")
        f.write(op)


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_json(annotation_files: List[dt.AnnotationFile]) -> Dict[str, Any]:
    categories: Dict[str, int] = calculate_categories(annotation_files)
    tag_categories: Dict[str, int] = calculate_tag_categories(annotation_files)
    return {
        "info": build_info(),
        "licenses": build_licenses(),
        "images": build_images(annotation_files, tag_categories),
        "annotations": list(build_annotations(annotation_files, categories)),
        "categories": list(build_categories(categories)),
        "tag_categories": list(build_tag_categories(tag_categories)),
    }


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def calculate_categories(annotation_files: List[dt.AnnotationFile]) -> Dict[str, int]:
    categories: Dict[str, int] = {}
    for annotation_file in annotation_files:
        for annotation_class in annotation_file.annotation_classes:
            if annotation_class.name not in categories and annotation_class.annotation_type in [
                "polygon",
                "complex_polygon",
                "bounding_box",
            ]:
                categories[annotation_class.name] = _calculate_category_id(annotation_class)
    return categories


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def calculate_tag_categories(annotation_files: List[dt.AnnotationFile]) -> Dict[str, int]:
    categories: Dict[str, int] = {}
    for annotation_file in annotation_files:
        for annotation_class in annotation_file.annotation_classes:
            if annotation_class.name not in categories and annotation_class.annotation_type == "tag":
                categories[annotation_class.name] = _calculate_category_id(annotation_class)
    return categories


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_info() -> Dict[str, Any]:
    # TODO fill out these fields in a meaningful way
    today = date.today()
    return {
        "description": "Exported from Darwin",
        "url": "n/a",
        "version": "n/a",
        "year": today.year,
        "contributor": "n/a",
        "date_created": today.strftime("%Y/%m/%d"),
    }


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_licenses() -> List[Dict[str, Any]]:
    return [{"url": "n/a", "id": 0, "name": "placeholder license"}]


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_images(annotation_files: List[dt.AnnotationFile], tag_categories: Dict[str, int]) -> List[Dict[str, Any]]:
    return [
        build_image(annotation_file, tag_categories)
        for annotation_file in sorted(annotation_files, key=lambda x: x.seq)
    ]


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_image(annotation_file: dt.AnnotationFile, tag_categories: Dict[str, int]) -> Dict[str, Any]:
    tags = [
        annotation for annotation in annotation_file.annotations if annotation.annotation_class.annotation_type == "tag"
    ]
    return {
        "license": 0,
        "file_name": annotation_file.filename,
        "coco_url": "n/a",
        "height": annotation_file.image_height,
        "width": annotation_file.image_width,
        "date_captured": "",
        "flickr_url": "n/a",
        "darwin_url": annotation_file.image_url,
        "darwin_workview_url": annotation_file.workview_url,
        "id": _build_image_id(annotation_file),
        "tag_ids": [tag_categories[tag.annotation_class.name] for tag in tags],
    }


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_annotations(
    annotation_files: List[dt.AnnotationFile], categories: Dict[str, int]
) -> Iterator[Optional[Dict[str, Any]]]:
    annotation_id = 0
    for annotation_file in annotation_files:
        for annotation in annotation_file.annotations:
            annotation_id += 1
            annotation_data = build_annotation(annotation_file, annotation_id, annotation, categories)
            if annotation_data:
                yield annotation_data


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_annotation(
    annotation_file: dt.AnnotationFile, annotation_id: int, annotation: dt.Annotation, categories: Dict[str, int]
) -> Optional[Dict[str, Any]]:
    annotation_type = annotation.annotation_class.annotation_type
    if annotation_type == "polygon":
        sequences = convert_polygons_to_sequences(annotation.data["path"], rounding=False)
        x_coords = [s[0::2] for s in sequences]
        y_coords = [s[1::2] for s in sequences]
        min_x = np.min([np.min(x_coord) for x_coord in x_coords])
        min_y = np.min([np.min(y_coord) for y_coord in y_coords])
        max_x = np.max([np.max(x_coord) for x_coord in x_coords])
        max_y = np.max([np.max(y_coord) for y_coord in y_coords])
        w = max_x - min_x
        h = max_y - min_y
        # Compute the area of the polygon
        poly_area = np.sum([polygon_area(x_coord, y_coord) for x_coord, y_coord in zip(x_coords, y_coords)])

        return {
            "id": annotation_id,
            "image_id": _build_image_id(annotation_file),
            "category_id": categories[annotation.annotation_class.name],
            "segmentation": sequences,
            "area": poly_area,
            "bbox": [min_x, min_y, w, h],
            "iscrowd": 0,
            "extra": build_extra(annotation),
        }
    elif annotation_type == "complex_polygon":
        mask = np.zeros((annotation_file.image_height, annotation_file.image_width))
        sequences = convert_polygons_to_sequences(annotation.data["paths"])
        draw_polygon(mask, sequences, 1)
        counts = rle_encode(mask)

        x_coords = [s[0::2] for s in sequences]
        y_coords = [s[1::2] for s in sequences]
        min_x = np.min([np.min(x_coord) for x_coord in x_coords])
        min_y = np.min([np.min(y_coord) for y_coord in y_coords])
        max_x = np.max([np.max(x_coord) for x_coord in x_coords])
        max_y = np.max([np.max(y_coord) for y_coord in y_coords])
        w = max_x - min_x + 1
        h = max_y - min_y + 1

        return {
            "id": annotation_id,
            "image_id": _build_image_id(annotation_file),
            "category_id": categories[annotation.annotation_class.name],
            "segmentation": {"counts": counts, "size": [annotation_file.image_height, annotation_file.image_width]},
            "area": np.sum(mask),
            "bbox": [min_x, min_y, w, h],
            "iscrowd": 1,
            "extra": build_extra(annotation),
        }
    elif annotation_type == "tag":
        pass
    elif annotation_type == "bounding_box":
        x = annotation.data["x"]
        y = annotation.data["y"]
        w = annotation.data["w"]
        h = annotation.data["h"]

        return build_annotation(
            annotation_file,
            annotation_id,
            dt.make_polygon(
                annotation.annotation_class.name,
                [{"x": x, "y": y}, {"x": x + w, "y": y}, {"x": x + w, "y": y + h}, {"x": x, "y": y + h}],
                None,
                annotation.subs,
            ),
            categories,
        )
    else:
        print(f"skipping unsupported annotation_type '{annotation_type}'")


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_extra(annotation: dt.Annotation) -> Dict[str, Any]:
    data = {}
    instance_id_sub = annotation.get_sub("instance_id")
    attributes_sub = annotation.get_sub("attributes")
    text_sub = annotation.get_sub("text")

    if instance_id_sub:
        data["instance_id"] = instance_id_sub.data
    if attributes_sub:
        data["attributes"] = attributes_sub.data
    if text_sub:
        data["text"] = text_sub.data
    return data


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_categories(categories: Dict[str, int]) -> Iterator[Dict[str, Any]]:
    for name, id in categories.items():
        yield {"id": id, "name": name, "supercategory": "root"}


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_tag_categories(categories: Dict[str, int]) -> Iterator[Dict[str, Any]]:
    for name, id in categories.items():
        yield {"id": id, "name": name}


@deprecation.deprecated(
    deprecated_in="0.7.7",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def polygon_area(x: np.ndarray, y: np.ndarray) -> float:
    """
    Returns the area of the input polygon, represented with two numpy arrays
    for x and y coordinates.
    """
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))


def _build_json(annotation_files: List[dt.AnnotationFile]) -> Dict[str, Any]:
    categories: Dict[str, int] = _calculate_categories(annotation_files)
    tag_categories: Dict[str, int] = _calculate_tag_categories(annotation_files)
    return {
        "info": _build_info(),
        "licenses": _build_licenses(),
        "images": _build_images(annotation_files, tag_categories),
        "annotations": list(_build_annotations(annotation_files, categories)),
        "categories": list(_build_categories(categories)),
        "tag_categories": list(_build_tag_categories(tag_categories)),
    }


def _calculate_categories(annotation_files: List[dt.AnnotationFile]) -> Dict[str, int]:
    categories: Dict[str, int] = {}
    for annotation_file in annotation_files:
        for annotation_class in annotation_file.annotation_classes:
            if annotation_class.name not in categories and annotation_class.annotation_type in [
                "polygon",
                "complex_polygon",
                "bounding_box",
            ]:
                categories[annotation_class.name] = _calculate_category_id(annotation_class)
    return categories


def _calculate_tag_categories(annotation_files: List[dt.AnnotationFile]) -> Dict[str, int]:
    categories: Dict[str, int] = {}
    for annotation_file in annotation_files:
        for annotation_class in annotation_file.annotation_classes:
            if annotation_class.name not in categories and annotation_class.annotation_type == "tag":
                categories[annotation_class.name] = _calculate_category_id(annotation_class)
    return categories


def _calculate_category_id(annotation_class: dt.AnnotationClass) -> int:
    return crc32(str.encode(annotation_class.name))


def _build_info() -> Dict[str, Any]:
    # TODO fill out these fields in a meaningful way
    today = date.today()
    return {
        "description": "Exported from Darwin",
        "url": "n/a",
        "version": "n/a",
        "year": today.year,
        "contributor": "n/a",
        "date_created": today.strftime("%Y/%m/%d"),
    }


def _build_licenses() -> List[Dict[str, Any]]:
    return [{"url": "n/a", "id": 0, "name": "placeholder license"}]


def _build_images(annotation_files: List[dt.AnnotationFile], tag_categories: Dict[str, int]) -> List[Dict[str, Any]]:
    return [
        _build_image(annotation_file, tag_categories)
        for annotation_file in sorted(annotation_files, key=lambda x: x.seq)
    ]


def _build_image(annotation_file: dt.AnnotationFile, tag_categories: Dict[str, int]) -> Dict[str, Any]:
    tags = [
        annotation for annotation in annotation_file.annotations if annotation.annotation_class.annotation_type == "tag"
    ]

    return {
        "license": 0,
        "file_name": annotation_file.filename,
        "coco_url": "n/a",
        "height": annotation_file.image_height,
        "width": annotation_file.image_width,
        "date_captured": "",
        "flickr_url": "n/a",
        "darwin_url": annotation_file.image_url,
        "darwin_workview_url": annotation_file.workview_url,
        "id": _build_image_id(annotation_file),
        "tag_ids": [tag_categories[tag.annotation_class.name] for tag in tags],
    }


def _build_image_id(annotation_file: dt.AnnotationFile) -> int:
    # CoCo file format requires unique image IDs
    # darwin 1.0 produces unique 'seq' values that can be used
    # darwin 2.0 does not provide `seq` so we hash the path + filename to produce a unique-enough 32bit int
    if annotation_file.seq:
        return annotation_file.seq
    else:
        full_path = str(Path(annotation_file.remote_path or "/") / Path(annotation_file.filename))
        return crc32(str.encode(full_path))


def _build_annotations(
    annotation_files: List[dt.AnnotationFile], categories: Dict[str, int]
) -> Iterator[Optional[Dict[str, Any]]]:
    annotation_id = 0
    for annotation_file in annotation_files:
        for annotation in annotation_file.annotations:
            annotation_id += 1
            annotation_data = _build_annotation(annotation_file, annotation_id, annotation, categories)
            if annotation_data:
                yield annotation_data


def _build_annotation(
    annotation_file: dt.AnnotationFile, annotation_id: int, annotation: dt.Annotation, categories: Dict[str, int]
) -> Optional[Dict[str, Any]]:
    annotation_type = annotation.annotation_class.annotation_type
    if annotation_type == "polygon":
        sequences = convert_polygons_to_sequences(annotation.data["path"], rounding=False)
        x_coords = [s[0::2] for s in sequences]
        y_coords = [s[1::2] for s in sequences]
        min_x = np.min([np.min(x_coord) for x_coord in x_coords])
        min_y = np.min([np.min(y_coord) for y_coord in y_coords])
        max_x = np.max([np.max(x_coord) for x_coord in x_coords])
        max_y = np.max([np.max(y_coord) for y_coord in y_coords])
        w = max_x - min_x
        h = max_y - min_y
        # Compute the area of the polygon
        poly_area = np.sum([_polygon_area(x_coord, y_coord) for x_coord, y_coord in zip(x_coords, y_coords)])

        return {
            "id": annotation_id,
            "image_id": _build_image_id(annotation_file),
            "category_id": categories[annotation.annotation_class.name],
            "segmentation": sequences,
            "area": poly_area,
            "bbox": [min_x, min_y, w, h],
            "iscrowd": 0,
            "extra": _build_extra(annotation),
        }
    elif annotation_type == "complex_polygon":
        mask = np.zeros((annotation_file.image_height, annotation_file.image_width))
        sequences = convert_polygons_to_sequences(annotation.data["paths"])
        draw_polygon(mask, sequences, 1)
        counts = rle_encode(mask)

        x_coords = [s[0::2] for s in sequences]
        y_coords = [s[1::2] for s in sequences]
        min_x = np.min([np.min(x_coord) for x_coord in x_coords])
        min_y = np.min([np.min(y_coord) for y_coord in y_coords])
        max_x = np.max([np.max(x_coord) for x_coord in x_coords])
        max_y = np.max([np.max(y_coord) for y_coord in y_coords])
        w = max_x - min_x + 1
        h = max_y - min_y + 1

        return {
            "id": annotation_id,
            "image_id": _build_image_id(annotation_file),
            "category_id": categories[annotation.annotation_class.name],
            "segmentation": {"counts": counts, "size": [annotation_file.image_height, annotation_file.image_width]},
            "area": np.sum(mask),
            "bbox": [min_x, min_y, w, h],
            "iscrowd": 1,
            "extra": _build_extra(annotation),
        }
    elif annotation_type == "tag":
        pass
    elif annotation_type == "bounding_box":
        x = annotation.data["x"]
        y = annotation.data["y"]
        w = annotation.data["w"]
        h = annotation.data["h"]

        return _build_annotation(
            annotation_file,
            annotation_id,
            dt.make_polygon(
                annotation.annotation_class.name,
                [{"x": x, "y": y}, {"x": x + w, "y": y}, {"x": x + w, "y": y + h}, {"x": x, "y": y + h}],
                None,
                annotation.subs,
            ),
            categories,
        )
    else:
        print(f"skipping unsupported annotation_type '{annotation_type}'")


def _build_extra(annotation: dt.Annotation) -> Dict[str, Any]:
    data = {}
    instance_id_sub = annotation.get_sub("instance_id")
    attributes_sub = annotation.get_sub("attributes")
    text_sub = annotation.get_sub("text")

    if instance_id_sub:
        data["instance_id"] = instance_id_sub.data
    if attributes_sub:
        data["attributes"] = attributes_sub.data
    if text_sub:
        data["text"] = text_sub.data
    return data


def _build_categories(categories: Dict[str, int]) -> Iterator[Dict[str, Any]]:
    for name, id in categories.items():
        yield {"id": id, "name": name, "supercategory": "root"}


def _build_tag_categories(categories: Dict[str, int]) -> Iterator[Dict[str, Any]]:
    for name, id in categories.items():
        yield {"id": id, "name": name}


def _polygon_area(x: np.ndarray, y: np.ndarray) -> float:
    """
    Returns the area of the input polygon, represented with two numpy arrays
    for x and y coordinates.
    """
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))



from typing import List

from pydantic import validator

from darwin.future.data_objects import validators as darwin_validators
from darwin.future.pydantic_base import DefaultDarwin


class Release(DefaultDarwin):
    """A class to manage all the information around a release on the darwin platform, including validation
    Attributes
    ----------
    name : str

    Methods
    ----------
    _name_validator: validates and auto formats the name variable
    """

    name: str

    # Data Validation
    _name_validator = validator("name", allow_reuse=True)(darwin_validators.parse_name)


ReleaseList = List[Release]



from typing import Any, Dict, List, Optional, Tuple, Union
from urllib import parse

from darwin.datatypes import ItemId


def inject_default_team_slug(method):
    """
    Injects team_slug if not specified
    """

    def wrapper(self, *args, **kwargs):
        if "team_slug" not in kwargs:
            kwargs["team_slug"] = self._default_team
        return method(self, *args, **kwargs)

    return wrapper


class BackendV2:
    def __init__(self, client: "Client", default_team):
        self._client = client
        self._default_team = default_team

    @inject_default_team_slug
    def register_data(
        self, dataset_slug: str, payload: Dict[str, Any], *, team_slug: Optional[str] = None
    ) -> Dict[str, Any]:

        payload["dataset_slug"] = dataset_slug
        response = self._client._post(
            endpoint=f"v2/teams/{team_slug}/items/register_upload",
            payload=payload,
            team_slug=team_slug,
        )
        return response

    @inject_default_team_slug
    def sign_upload(self, dataset_slug: str, upload_id: str, *, team_slug: Optional[str] = None) -> Dict[str, Any]:
        return self._client._get(f"v2/teams/{team_slug}/items/uploads/{upload_id}/sign", team_slug=team_slug)

    @inject_default_team_slug
    def confirm_upload(self, dataset_slug: str, upload_id: str, *, team_slug: Optional[str] = None) -> Dict[str, Any]:
        return self._client._post(
            f"v2/teams/{team_slug}/items/uploads/{upload_id}/confirm",
            payload={},
            team_slug=team_slug,
        )

    @inject_default_team_slug
    def fetch_items(
        self, dataset_id: int, cursor: Union[Dict[str, Any], List[Tuple[str, Any]]], *, team_slug: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Fetch the remote items from the given dataset.

        Parameters
        ----------
        dataset_id: int
            Id of the dataset the file belong to.
        cursor: Dict[str, Any]
            Number of items per page and page number. Defaults to {"page[size]": 500, "page[from]": 0}.
        payload: Dict[str, Any]
            Filter and sort parameters.
        team_slug: str
            The team slug of the dataset.

        Returns
        -------
         Dict[str, Any]
            A response dictionary with the file information.
        """
        if isinstance(cursor, dict):
            cursor = list(cursor.items())

        cursor.append(("dataset_ids[]", dataset_id))

        return self._client._get(f"/v2/teams/{team_slug}/items?{parse.urlencode(cursor, True)}", team_slug)

    @inject_default_team_slug
    def archive_items(self, payload: Dict[str, Any], *, team_slug: Optional[str] = None) -> None:
        """
        Archives the item from the given dataset.

        Parameters
        ----------
        team_slug: str
            The slug of the team.
        payload: Dict[str, Any]
            A filter Dictionary that defines the items to be archived.
        """
        self._client._post(f"v2/teams/{team_slug}/items/archive", payload, team_slug)

    @inject_default_team_slug
    def restore_archived_items(self, payload: Dict[str, Any], *, team_slug: Optional[str] = None) -> None:
        """
        Restores the archived item from the given dataset.

        Parameters
        ----------
        team_slug: str
            The slug of the team.
        payload: Dict[str, Any]
            A filter Dictionary that defines the items to be restored.
        """
        self._client._post(f"v2/teams/{team_slug}/items/restore", payload, team_slug)

    @inject_default_team_slug
    def move_to_stage(
        self, filters: Dict[str, Any], stage_id: str, workflow_id: str, *, team_slug: Optional[str] = None
    ) -> None:
        """
        Moves the given items to the specified stage

        Parameters
        ----------
        dataset_slug: str
            The slug of the dataset.
        team_slug: str
            The slug of the team.
        payload: Dict[str, Any]
            A filter Dictionary that defines the items to have the 'new' status.
        """
        payload = {"filters": filters, "stage_id": stage_id, "workflow_id": workflow_id}
        self._client._post_raw(f"v2/teams/{team_slug}/items/stage", payload, team_slug)

    @inject_default_team_slug
    def get_dataset(self, id: str, *, team_slug: Optional[str] = None) -> Dict[str, Any]:
        return self._client._get(f"datasets/{id}", team_slug)

    @inject_default_team_slug
    def get_workflow(self, id: str, *, team_slug: Optional[str] = None) -> Dict[str, Any]:
        return self._client._get(f"v2/teams/{team_slug}/workflows/{id}", team_slug)

    @inject_default_team_slug
    def delete_items(self, filters, *, team_slug: Optional[str] = None):
        self._client._delete(f"v2/teams/{team_slug}/items", {"filters": filters}, team_slug)

    @inject_default_team_slug
    def export_dataset(
        self,
        name,
        format,
        include_authorship,
        include_token,
        dataset_slug,
        filters,
        annotation_class_ids,
        *,
        team_slug: Optional[str] = None,
    ):
        payload = {
            "include_authorship": include_authorship,
            "include_export_token": include_token,
            "name": name,
            "annotation_filters": {},
        }
        if format:
            payload["format"] = format

        if annotation_class_ids:
            payload["annotation_filters"] = {"annotation_class_ids": list(map(int, annotation_class_ids))}
        if filters is not None:
            # Backend assumes default filters only if those are completely missing.
            payload["filters"] = filters

        return self._client._post(f"v2/teams/{team_slug}/datasets/{dataset_slug}/exports", payload, team_slug)

    def get_exports(self, dataset_slug, *, team_slug: Optional[str] = None):
        return self._client._get(f"v2/teams/{team_slug}/datasets/{dataset_slug}/exports", team_slug)

    @inject_default_team_slug
    def post_comment(self, item_id, text, x, y, w, h, slot_name, team_slug: Optional[str] = None):
        payload = {
            "bounding_box": {"h": h, "w": w, "x": x, "y": y},
            "comments": [{"body": text}],
            "slot_name": slot_name,
        }
        return self._client._post(f"v2/teams/{team_slug}/items/{item_id}/comment_threads", payload, team_slug)

    @inject_default_team_slug
    def import_annotation(self, item_id: ItemId, payload: Dict[str, Any], team_slug: Optional[str] = None) -> None:
        """
        Imports the annotation for the item with the given id.

        Parameters
        ----------
        item_id: ItemId
            Identifier of the Item that we are import the annotation to.
        payload: Dict[str, Any]
            A dictionary with the annotation to import. The default format is:
            `{"annotations": serialized_annotations, "overwrite": "false"}`
        """

        return self._client._post_raw(f"v2/teams/{team_slug}/items/{item_id}/import", payload=payload)



from typing import Generator, List

from darwin.datatypes import UnknownType


def flatten_list(list_of_lists: List[UnknownType]) -> List[UnknownType]:
    """
    Flattens a list of lists into a single list.

    Parameters
    ----------
    list_of_lists : List[List[Any]]
        The list of lists to flatten.

    Returns
    -------
    List[Any]
        The flattened list.
    """

    if not isinstance(list_of_lists, list):
        raise TypeError("Expected a list")

    def flatten(lists: List[UnknownType]) -> Generator[list, UnknownType, UnknownType]:
        if isinstance(lists, list) and len(lists) == 0:
            return lists
        for item in lists:
            if isinstance(item, list):
                for i in flatten(item):
                    yield i
            else:
                yield item

    return list(flatten(list_of_lists))



from pathlib import Path
from typing import Iterable

import darwin.datatypes as dt
from darwin.exporter.formats.mask import export as export_mask


def export(annotation_files: Iterable[dt.AnnotationFile], output_dir: Path) -> None:
    """
    Exports the given ``AnnotationFile``\\s into semantic masks inside of the given ``output_dir``.

    Parameters
    ----------
    annotation_files : Iterable[dt.AnnotationFile]
        The ``AnnotationFile``\\s to be exported.
    output_dir : Path
        The folder where the new semantic mask files will be.
    """
    return export_mask(annotation_files=annotation_files, output_dir=output_dir, mode="rgb")



def parse_name(name: str) -> str:
    """
    A function to parse and validate a name

    Parameters
    ----------
    name : str
        The name to be parsed and validated

    Returns
    -------
    str
        The parsed and validated name
    """
    assert isinstance(name, str)
    return name.lower().strip()



class_map = {"type": "object", "patternProperties": {"^([0-9]+)+$": {"type": "string"}}, "additionalProperties": False}

nifti_image_label_pair = {
    "type": "object",
    "properties": {
        "image": {"type": "string"},
        "label": {"type": "string"},
        "class_map": class_map,
        "mode": {"type": "string", "enum": ["image", "video", "instances"]},
    },
    "required": ["image", "label", "class_map"],
    "additionalProperties": False,
}

nifti_import_schema = {
    "type": "object",
    "properties": {"data": {"type": "array", "items": nifti_image_label_pair}},
    "required": ["data"],
    "additionalProperties": False,
}



from functools import partial, reduce
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, cast

import orjson as json
from jsonschema import validate

from darwin.datatypes import (
    Annotation,
    AnnotationClass,
    AnnotationFile,
    Point,
    SubAnnotation,
    make_bounding_box,
    make_keypoint,
    make_line,
    make_polygon,
    make_tag,
)
from darwin.importer.formats.labelbox_schemas import labelbox_export
from darwin.utils import attempt_decode


def parse_path(path: Path) -> Optional[List[AnnotationFile]]:
    """
    Parses the given LabelBox file and maybe returns the corresponding annotations.
    The file must have a structure similar to the following:

    .. code-block:: javascript

        [
            {
                "Label":{
                    "objects":[
                        {
                            "title": "SomeTitle",
                            "bbox":{"top":3558, "left":145, "height":623, "width":449}
                        },
                        { }
                    ],
                    "classifications": [
                        {
                            "value": "a_question",
                            "answer": {"value": "an_answer"}
                        }
                    ]
                },
                "External ID": "demo-image-7.jpg"
            },
            { }
        ]

    You can check the Labelbox Schemas in `labelbox_schemas.py`.

    Currently we support the following annotations:

        - bounding-box ``Image``: https://docs.labelbox.com/docs/bounding-box-json
        - polygon ``Image``: https://docs.labelbox.com/docs/polygon-json
        - point ``Image``: https://docs.labelbox.com/docs/point-json
        - polyline ``Image``: https://docs.labelbox.com/docs/polyline-json

    We also support conversion from question/answer to Annotation Tags for the following:

        - Radio Buttons
        - Checklists
        - Free Text

    Parameters
    --------
    path: Path
        The path of the file to parse.

    Returns
    -------
    Optional[List[darwin.datatypes.AnnotationFile]]
        The AnnotationFiles with the parsed information from the file or None, if the file is not a
        `json` file.

    Raises
    ------
    ValidationError
        If the given JSON file is malformed or if it has an unknown annotation.
        To see a list of possible annotation formats go to:
        https://docs.labelbox.com/docs/annotation-types-1

    """
    if path.suffix != ".json":
        return None
    data = attempt_decode(path)
    validate(data, schema=labelbox_export)
    convert_with_path = partial(_convert, path=path)

    return _map_list(convert_with_path, data)


def _convert(file_data: Dict[str, Any], path) -> AnnotationFile:
    filename: str = str(file_data.get("External ID"))
    label: Dict[str, Any] = cast(Dict[str, Any], file_data.get("Label"))
    label_objects: List[Dict[str, Any]] = cast(List[Dict[str, Any]], label.get("objects"))
    label_classifications: List[Dict[str, Any]] = cast(List[Dict[str, Any]], label.get("classifications"))

    classification_annotations: List[Annotation] = []
    if len(label_classifications) > 0:
        classification_annotations = _flat_map_list(_map_list(_convert_label_classifications, label_classifications))

    object_annotations: List[Annotation] = _map_list(_convert_label_objects, label_objects)
    annotations: List[Annotation] = object_annotations + classification_annotations

    classes: Set[AnnotationClass] = set(map(_get_class, annotations))
    return AnnotationFile(
        annotations=annotations, path=path, filename=filename, annotation_classes=classes, remote_path="/"
    )


def _convert_label_objects(obj: Dict[str, Any]) -> Annotation:
    title: str = str(obj.get("title"))
    bbox: Optional[Dict[str, Any]] = obj.get("bbox")
    if bbox:
        return _to_bbox_annotation(bbox, title)

    polygon: Optional[List[Point]] = obj.get("polygon")
    if polygon:
        return _to_polygon_annotation(polygon, title)

    point: Optional[Point] = obj.get("point")
    if point:
        return _to_keypoint_annotation(point, title)

    line: Optional[List[Point]] = obj.get("line")
    if line:
        return _to_line_annotation(line, title)

    raise ValueError(f"Unknown label object {obj}")


def _convert_label_classifications(obj: Dict[str, Any]) -> List[Annotation]:
    question: str = str(obj.get("value"))

    answer: Optional[Dict[str, Any]] = obj.get("answer")
    if answer is not None:
        if isinstance(answer, str):
            return [_to_tag_annotations_from_free_text(question, answer)]
        else:
            return [_to_tag_annotations_from_radio_box(question, answer)]

    answers: Optional[List[Dict[str, Any]]] = obj.get("answers")
    if answers is not None:
        return _to_tag_annotations_from_checklist(question, answers)

    raise ValueError(f"Unknown classification obj {obj}")


def _to_bbox_annotation(bbox: Dict[str, Any], title: str) -> Annotation:
    x: float = cast(float, bbox.get("left"))
    y: float = cast(float, bbox.get("top"))
    width: float = cast(float, bbox.get("width"))
    height: float = cast(float, bbox.get("height"))

    return make_bounding_box(title, x, y, width, height)


def _to_polygon_annotation(polygon: List[Point], title: str) -> Annotation:
    return make_polygon(title, polygon, None)


def _to_keypoint_annotation(point: Point, title: str) -> Annotation:
    x: float = cast(float, point.get("x"))
    y: float = cast(float, point.get("y"))

    return make_keypoint(title, x, y)


def _to_line_annotation(line: List[Point], title: str) -> Annotation:
    return make_line(title, line, None)


def _to_tag_annotations_from_radio_box(question: str, radio_button: Dict[str, Any]) -> Annotation:
    answer: str = str(radio_button.get("value"))
    return make_tag(f"{question}:{answer}")


def _to_tag_annotations_from_checklist(question: str, checklist) -> List[Annotation]:
    annotations: List[Annotation] = []
    for answer in checklist:
        val: str = answer.get("value")
        annotations.append(make_tag(f"{question}:{val}"))

    return annotations


def _to_tag_annotations_from_free_text(question: str, free_text: str) -> Annotation:
    return make_tag(question, [SubAnnotation(annotation_type="text", data=free_text)])


def _get_class(annotation: Annotation) -> AnnotationClass:
    return annotation.annotation_class


def _flat_map_list(the_list: List[List[Any]]) -> List[Any]:
    # We do a flat_map here: https://stackoverflow.com/a/2082107/1337392
    return reduce(list.__add__, the_list)


def _map_list(fun: Callable[[Any], Any], the_list: List[Any]) -> List[Any]:
    return list(map(fun, the_list))



import csv
from pathlib import Path
from typing import List, Optional

import darwin.datatypes as dt
from darwin.path_utils import deconstruct_full_path


def parse_path(path: Path) -> Optional[List[dt.AnnotationFile]]:
    """
    Parses the given file and returns a ``List[dt.AnnotationFile]`` with the parsed files, or
    ``None`` if the given file's extension is not ``.csv``.

    Parameters
    ----------
    path : Path
        The ``Path`` of the file to parse.

    Returns
    -------
    Optional[List[dt.AnnotationFile]]
        A ``List[dt.AnnotationFile]`` or ``None`` if the function was not able to parse the file.
    """
    if path.suffix != ".csv":
        return None

    files = []
    with path.open() as f:
        reader = csv.reader(f)
        for row in reader:
            filename, *tags = map(lambda s: s.strip(), row)
            if filename == "":
                continue
            annotations = [dt.make_tag(tag) for tag in tags if len(tag) > 0]
            annotation_classes = set([annotation.annotation_class for annotation in annotations])
            remote_path, filename = deconstruct_full_path(filename)
            files.append(dt.AnnotationFile(path, filename, annotation_classes, annotations, remote_path=remote_path))
    return files



from pydantic import BaseModel


class DefaultDarwin(BaseModel):
    """Default Darwin-Py pydantic settings for meta information.
    Default settings include:
        - auto validating variables on setting/assignment
        - underscore attributes are private
        - objects are passed by reference to prevent unnecesary data copying
    """

    class Config:
        validate_assignment = True
        underscore_attrs_are_private = True
        copy_on_model_validation = "none"



import os
import shutil
from pathlib import Path
from typing import Iterable

import numpy as np
from PIL import Image

import darwin.datatypes as dt
from darwin.utils import convert_polygons_to_mask, get_progress_bar, ispolygon


def export(annotation_files: Iterable[dt.AnnotationFile], output_dir: Path) -> None:
    """
    Exports the given ``AnnotationFile``\\s into instance masks format inside of the given
    ``output_dir``. Deletes everything within ``output_dir/masks`` before writting to it.

    Parameters
    ----------
    annotation_files : Iterable[dt.AnnotationFile]
        The ``AnnotationFile``\\s to be exported.
    output_dir : Path
        The folder where the new instance mask files will be.
    """
    masks_dir = output_dir / "masks"
    if masks_dir.exists():
        shutil.rmtree(masks_dir)
    masks_dir.mkdir(parents=True, exist_ok=True)
    with open(output_dir / "instance_mask_annotations.csv", "w") as f:
        f.write("image_id,mask_id,class_name\n")
        for annotation_file in get_progress_bar(list(annotation_files), "Processing annotations"):
            image_id = os.path.splitext(annotation_file.filename)[0]
            height = annotation_file.image_height
            width = annotation_file.image_width
            annotations = [a for a in annotation_file.annotations if ispolygon(a.annotation_class)]
            for i, annotation in enumerate(annotations):
                cat = annotation.annotation_class.name
                if annotation.annotation_class.annotation_type == "polygon":
                    polygon = annotation.data["path"]
                elif annotation.annotation_class.annotation_type == "complex_polygon":
                    polygon = annotation.data["paths"]
                else:
                    continue
                mask = convert_polygons_to_mask(polygon, height=height, width=width, value=255)
                mask = Image.fromarray(mask.astype(np.uint8))
                mask_id = f"{image_id}_{i:05}"
                outfile = masks_dir / f"{mask_id}.png"
                outfile.parent.mkdir(parents=True, exist_ok=True)
                mask.save(outfile)
                f.write(f"{image_id},{mask_id},{cat}\n")



import re
from typing import Optional, Tuple, Union


class DatasetIdentifier:
    """
    Formal representation of a dataset identifier for the SDK.

    A dataset identifier is a string that uniquely identifies a dataset on Darwin.
    A dataset identifier is made of the following substrings:
    ``<team-slug>/<dataset-slug>:<version>``.

    If ``version`` is missing, it defaults to ``latest``.

    Parameters
    ----------
    dataset_slug : str
        The slugified name of the dataset.
    team_slug : Optional[str], default: None
        The slugified name of the team.
    version : Optional[str], default: None
        The version of the identifier.

    Attributes
    ----------
    dataset_slug : str
        The slugified name of the dataset.
    team_slug : Optional[str], default: None
        The slugified name of the team.
    version : Optional[str], default: None
        The version of the identifier.
    """

    def __init__(self, dataset_slug: str, team_slug: Optional[str] = None, version: Optional[str] = None):
        self.dataset_slug = dataset_slug
        self.team_slug = team_slug
        self.version = version

    @classmethod
    def parse(cls, identifier: Union[str, "DatasetIdentifier"]) -> "DatasetIdentifier":
        """
        Parses the given identifier and returns the corresponding DatasetIdentifier.

        Parameters
        ----------
        identifier : Union[str, DatasetIdentifier]
            The identifier to be parsed.

        Returns
        -------
        DatasetIdentifier
            The SDK representation of a ``DatasetIdentifier``.

        Raises
        ------
        ValueError
            If the ``identifier`` given is invalid.
        """
        if isinstance(identifier, DatasetIdentifier):
            return identifier

        team_slug, dataset_slug, version = _parse(identifier)
        return cls(dataset_slug=dataset_slug, team_slug=team_slug, version=version)

    def __str__(self):
        output = ""
        if self.team_slug:
            output = f"{self.team_slug}/"
        output = f"{output}{self.dataset_slug}"
        if self.version:
            output = f"{output}:{self.version}"
        return output


def _parse(slug: str) -> Tuple[Optional[str], str, Optional[str]]:
    team: Optional[str] = None
    version: Optional[str] = None

    if not _is_slug_valid(slug):
        raise ValueError(f"Invalid dataset identifier {slug}")

    initial_split = slug.split("/")
    if len(initial_split) == 1:
        dataset = initial_split[0]
    elif len(initial_split) == 2:
        team, dataset = initial_split
    else:
        raise ValueError(f"Invalid dataset identifier {slug}")

    if ":" in dataset:
        dataset, version = dataset.split(":")

    return team, dataset, version


def _is_slug_valid(slug: str) -> bool:
    slug_format = "[\\_a-zA-Z0-9.-]+"
    version_format = "[\\_a-zA-Z0-9.:-]+"
    return re.fullmatch(rf"({slug_format}/)?{slug_format}(:{version_format})?", slug) is not None



bounding_box = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/bounding_box",
    "description": "Schema of a Bounding Box",
    "title": "Bounding Box",
    "default": {"top": 1.2, "left": 2.5, "height": 10, "width": 20},
    "examples": [{"top": 0, "left": 0, "height": 10, "width": 20}],
    "type": "object",
    "properties": {
        "top": {"type": "number"},
        "left": {"type": "number"},
        "height": {"type": "number", "minimum": 0},
        "width": {"type": "number", "minimum": 0},
    },
    "required": ["top", "left", "height", "width"],
}

point = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/point",
    "description": "Schema of a Point",
    "title": "Point",
    "default": {"x": 1.2, "y": 2.5},
    "examples": [{"x": 1.2, "y": 2.5}, {"x": 0, "y": 1}],
    "type": "object",
    "properties": {"x": {"type": "number"}, "y": {"type": "number"}},
    "required": ["x", "y"],
}

polygon = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/polygon",
    "description": "Schema of a Polygon",
    "title": "Polygon",
    "default": [{"x": 1.2, "y": 2.5}, {"x": 2.5, "y": 3.6}, {"x": 1.2, "y": 2.5}],
    "examples": [[{"x": 1.2, "y": 2.5}, {"x": 2.5, "y": 3.6}, {"x": 1.2, "y": 2.5}], []],
    "type": "array",
    "items": point,
}

polyline = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/polyline",
    "description": "Schema of a Polyline",
    "title": "Polyline",
    "default": [{"x": 1.2, "y": 2.5}, {"x": 2.5, "y": 3.6}, {"x": 1.2, "y": 2.5}],
    "examples": [[{"x": 1.2, "y": 2.5}, {"x": 2.5, "y": 3.6}, {"x": 1.2, "y": 2.5}], []],
    "type": "array",
    "items": point,
}

label_object = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/label_object",
    "description": "An object belonging to the objects array from a Label",
    "title": "Label Object",
    "default": {"title": "Banana", "point": {"x": 3665.814, "y": 351.628}},
    "examples": [
        {"title": "Banana", "point": {"x": 3665.814, "y": 351.628}},
        {"title": "Orange", "bbox": {"top": 1.2, "left": 2.5, "height": 10, "width": 20}},
        {"title": "Apple", "polygon": [{"x": 1.2, "y": 2.5}, {"x": 2.5, "y": 3.6}, {"x": 1.2, "y": 2.5}]},
    ],
    "type": "object",
    "required": ["title"],
    "properties": {"title": {"type": "string"}},
    "oneOf": [
        {"required": ["point"], "properties": {"point": point}},
        {"required": ["bbox"], "properties": {"bbox": bounding_box}},
        {"required": ["polygon"], "properties": {"polygon": polygon}},
        {"required": ["line"], "properties": {"line": polyline}},
    ],
}

classification_answer_obj = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/classification_answer_obj",
    "description": "Schema of a Classification Answer Object",
    "title": "Answer",
    "default": {"value": "an_answer"},
    "examples": [{"value": "an_answer"}],
    "type": "object",
    "properties": {"value": {"type": "string"}},
    "required": ["value"],
}

classification_answer_free_text = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/classification_answer_free_text",
    "description": "Schema of a Classification Free Text Answer",
    "title": "Free Text Answer",
    "default": "A free text answer",
    "examples": ["A free text answer"],
    "type": "string",
}

label_classification = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/label_classification",
    "description": "A classification belonging to the classifications array from a Label",
    "title": "Label Classification",
    "default": {"value": "a_question", "answer": {"value": "an_answer"}},
    "examples": [
        {"value": "a_question", "answer": {"value": "an_answer"}},
        {"value": "a_question", "answers": [{"value": "an_answer_1"}, {"value": "an_answer_2"}]},
    ],
    "type": "object",
    "required": ["value"],
    "properties": {"value": {"type": "string"}},
    "oneOf": [
        {
            "required": ["answer"],
            "properties": {"answer": {"oneOf": [classification_answer_free_text, classification_answer_obj]}},
        },
        {"required": ["answers"], "properties": {"answers": {"type": "array", "items": classification_answer_obj}}},
    ],
}

labelbox_file = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/label_file",
    "description": "A Labelbox file, equivalent to a Darwin AnnotationFile",
    "title": "Labelbox File",
    "default": {
        "Label": {
            "objects": [
                {
                    "title": "Banana",
                    "point": {"x": 3665.814, "y": 351.628},
                }
            ]
        },
        "External ID": "demo-image-7.jpg",
    },
    "examples": [
        {
            "Label": {
                "objects": [
                    {
                        "title": "Banana",
                        "point": {"x": 3665.814, "y": 351.628},
                    }
                ]
            },
            "External ID": "demo-image-7.jpg",
        }
    ],
    "type": "object",
    "properties": {
        "Label": {
            "type": "object",
            "properties": {
                "objects": {"type": "array", "items": label_object},
                "classifications": {"type": "array", "items": label_classification},
            },
            "required": ["objects", "classifications"],
        },
        "External ID": {"type": "string"},
    },
    "required": ["Label", "External ID"],
}

labelbox_export = {
    "$id": "https://darwin.v7labs.com/schemas/labelbox/export",
    "description": "A Labelbox export",
    "title": "Export",
    "default": [
        {
            "Label": {
                "objects": [
                    {
                        "title": "Banana",
                        "point": {"x": 3665.814, "y": 351.628},
                    }
                ]
            },
            "External ID": "demo-image-7.jpg",
        }
    ],
    "examples": [
        [
            {
                "Label": {
                    "objects": [
                        {
                            "title": "Banana",
                            "point": {"x": 3665.814, "y": 351.628},
                        }
                    ]
                },
                "External ID": "demo-image-7.jpg",
            },
            {
                "Label": {
                    "objects": [
                        {
                            "title": "Orange",
                            "point": {"x": 0.814, "y": 0.628},
                        }
                    ]
                },
                "External ID": "demo-image-8.jpg",
            },
        ]
    ],
    "type": "array",
    "items": labelbox_file,
}



import csv
from pathlib import Path
from typing import Dict, List, Optional, Union

import darwin.datatypes as dt


def parse_path(path: Path) -> Optional[List[dt.AnnotationFile]]:
    """
    Parses the given ``csv video`` file and returns a ``List[dt.AnnotationFile]``, or ``None`` if
    the extension of the given file is not ``.csv``.

    Parameters
    ----------
    path : Path
        The ``Path`` to the file to be parsed.

    Returns
    -------
    Optional[List[dt.AnnotationFile]]
        A ``List[dt.AnnotationFile]`` or ``None`` if the file was parseable.
    """
    if path.suffix != ".csv":
        return None

    files = []

    file_annotation_map: Dict[str, List[Union[dt.Annotation, dt.VideoAnnotation]]] = {}
    with path.open() as f:
        reader = csv.reader(f)
        for row in reader:
            try:
                filename, tag, start_frame, end_frame = map(lambda s: s.strip(), row)
            except ValueError:
                continue
            if filename == "":
                continue

            start_frame = int(start_frame)
            end_frame = int(end_frame)

            annotation = dt.make_tag(tag)
            frames = {i: annotation for i in range(start_frame, end_frame + 1)}
            keyframes = {i: i == start_frame for i in range(start_frame, end_frame + 1)}

            annotation = dt.make_video_annotation(frames, keyframes, [[start_frame, end_frame]], False, slot_names=[])
            if filename not in file_annotation_map:
                file_annotation_map[filename] = []
            file_annotation_map[filename].append(annotation)
    for filename in file_annotation_map:
        annotations = file_annotation_map[filename]
        annotation_classes = set([annotation.annotation_class for annotation in annotations])
        files.append(dt.AnnotationFile(path, filename, annotation_classes, annotations, is_video=True, remote_path="/"))
    return files



from pathlib import Path
from typing import Iterable

import darwin.datatypes as dt
from darwin.exporter.formats.mask import export as export_mask


def export(annotation_files: Iterable[dt.AnnotationFile], output_dir: Path) -> None:
    return export_mask(annotation_files=annotation_files, output_dir=output_dir, mode="grey")



import unittest

import pytest
from pydantic import BaseModel, ValidationError

from darwin.future.data_objects.dataset import Dataset
from darwin.future.data_objects.release import Release
from darwin.future.data_objects.team import Team


@pytest.fixture
def basic_team() -> dict:
    return {"slug": "test-team", "id": 0}


@pytest.fixture
def basic_dataset() -> dict:
    return {"name": "test-dataset", "slug": "test-dataset"}


@pytest.fixture
def basic_release() -> dict:
    return {"name": "test-release"}


@pytest.fixture
def basic_combined(basic_team: dict, basic_dataset: dict, basic_release: dict) -> dict:
    combined = basic_team
    combined["datasets"] = [basic_dataset]
    combined["datasets"][0]["releases"] = [basic_release]
    return combined


@pytest.fixture
def broken_combined(basic_combined: dict) -> dict:
    del basic_combined["datasets"][0]["name"]
    return basic_combined


def test_integrated_parsing_works_with_raw(basic_combined: dict) -> None:
    team = Team.parse_obj(basic_combined)
    assert team.slug == "test-team"
    assert team.datasets is not None
    assert team.datasets[0].name == "test-dataset"
    assert team.datasets[0].releases is not None
    assert team.datasets[0].releases[0].name == "test-release"


def test_broken_obj_raises(broken_combined: dict) -> None:
    with pytest.raises(ValidationError) as e_info:
        broken = Team.parse_obj(broken_combined)


@pytest.mark.parametrize("test_object", [Team, Dataset, Release])
def test_empty_obj_raises(test_object: BaseModel) -> None:
    with pytest.raises(ValidationError) as e_info:
        broken = test_object.parse_obj({})



import io
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import yaml

from darwin.datatypes import PathLike, Team
from darwin.exceptions import InvalidCompressionLevel, InvalidTeam


class Config(object):
    """Handles YAML configuration files"""

    def __init__(self, path: Optional[PathLike] = None):
        """
        If path is None the config will be in memory only
        """
        if isinstance(path, str):
            path = Path(path)

        self._path: Optional[Path] = path
        self._data: Dict[str, Any] = self._parse()

    def _parse(self) -> Dict[str, Any]:
        """Parses the YAML configuration file"""
        if not self._path:
            return {}
        try:
            with open(self._path, "r") as stream:
                return yaml.safe_load(stream)
        except FileNotFoundError:
            return {}

    def get(self, key: Union[str, List[str]], default: Optional[Any] = None) -> Any:
        """
        Gets the value defined by key.

        Parameters
        ----------
        key: Union[str, List[str]]
            The key where the value to be fetched is stored.
            It can be formatted as a simple string, or as a path/like/string to fetch nested values.
        default: Optional[Any]
            A default value in case the given key returns ``None``. Defaults to ``None``.

        Returns
        -------
        Any
            The value stored by the key.
        """

        acc: Any = self._data.copy()

        while True:
            if isinstance(key, str):
                key = key.split("/")
            key, *keys = key
            acc = acc.get(key)
            if acc is None:
                return default
            if len(keys) == 0:
                return acc
            else:
                key = keys

    def put(self, key: Union[str, List[str]], value: Any, save: bool = True) -> None:
        """
        Sets the value for the specified key.

        Parameters
        ----------
        key: Union[str, List[str]]
            The key where the value is going to be stored.
            It can be formatted as a simple string, or as a path/like/string to fetch nested values.
        value: Any
            The value to be stored.
        save: bool
            If ``True``, persists the value in the FileSystem. Defaults to ``True``.
        """
        if isinstance(key, str):
            key = key.split("/")

        pointer = self._data

        for k in key[:-1]:
            pointer = pointer.setdefault(k, {})
        pointer[key[-1]] = str(value)

        if save:
            self._save()

    def _save(self) -> None:
        """Persist the configuration to the file system"""
        if not self._path:
            return
        with io.open(self._path, "w", encoding="utf8") as f:
            yaml.dump(self._data, f, default_flow_style=False, allow_unicode=True)

    def set_team(self, team: str, api_key: str, datasets_dir: str) -> None:
        """
        Stores the API key and Dataset directory for the given team.

        Parameters
        ----------
        team: str
            The name of the team.
        api_key: str
            The API key the user has to do actions in the given team.
        datasets_dir: str
            The directory to stores datasets from the given team.
        """
        self.put(f"teams/{team}/api_key", api_key)
        self.put(f"teams/{team}/datasets_dir", datasets_dir)

    def set_default_team(self, team: str) -> None:
        """
        Sets the given team as the default one.

        Parameters
        ----------
        team: str
            The team's slug.

        Raises
        ------
        InvalidTeam
            If the given team is not in the configuration file. Authenticate with this team first
            to avoid this issue.
        """
        if self.get(f"teams/{team}") is None:
            raise InvalidTeam()
        self.put("global/default_team", team)

    def set_compression_level(self, level: int) -> None:
        """
        Sets the given compression level globally.

        Parameters
        ----------
        level: int
            The compression level.

        Raises
        ------
        InvalidCompressionLevel
            Compression level is out of supported range. Use number from 0 to 9 to avoid this issue.
        """
        if level < 0 or level > 9:
            raise InvalidCompressionLevel(level)
        self.put("global/payload_compression_level", level)

    def set_global(self, api_endpoint: str, base_url: str, default_team: Optional[str] = None) -> None:
        """
        Stores the url to access teams. If a default team is given, it also stores that team as the
        globaly default one.

        Parameters
        ----------
        api_endpoint: str
            The '/api' endpoint from V7's API.
        base_url: str
            The base URL for V7 together with protocol.
        default_team: Optional[str]
            The default team's slug. Defaults to ``None``.
        """
        self.put("global/api_endpoint", api_endpoint)
        self.put("global/base_url", base_url)
        if default_team:
            self.put("global/default_team", default_team)

    def get_team(self, team: Optional[str] = None, raise_on_invalid_team: bool = True) -> Optional[Team]:
        """
        Returns the Team object from the team with the given slug if an authentication with an API
        key was performed earlier.

        Parameters
        ----------
        team: Optional[str]
            The Team's slug. If none is given, searches for the default team instead. Defaults to
            ``None``.
        raise_on_invalid_team: bool
            If ``True``, raises if no team is found, if False returns ``None`` instead. Defaults to ``True``.

        Returns
        -------
        Optional[Team]:
            The team or ``None`` if no API key for the team was found and `raise_on_invalid_team` is
            ``False``.

        Raises
        ------
        InvalidTeam
            If the user has not been authenticated with an API key for they given team.
        """
        if not team:
            return self.get_default_team(raise_on_invalid_team=raise_on_invalid_team)

        api_key = self.get(f"teams/{team}/api_key")
        if api_key is None:
            if raise_on_invalid_team:
                raise InvalidTeam()
            else:
                return None
        default: bool = self.get("global/default_team") == team or len(list(self.get("teams").keys())) == 1

        datasets_dir = self.get(f"teams/{team}/datasets_dir")
        return Team(slug=team, api_key=api_key, default=default, datasets_dir=datasets_dir)

    def get_default_team(self, raise_on_invalid_team: bool = True) -> Optional[Team]:
        """
        Returns the default Team if one is defined.

        Parameters
        ----------
        raise_on_invalid_team: bool
            If ``True``, raises if no default team is found, if False returns ``None`` instead. Defaults
            to ``True``.

        Returns
        -------
        Optional[Team]:
            The team or ``None`` if no default team is set and `raise_on_invalid_team` is ``False``.

        Raises
        ------
        InvalidTeam
            If the user has not set a default team.
        """
        default_team = self.get("global/default_team")
        if default_team:
            return self.get_team(default_team)
        teams = list((self.get("teams") or {}).keys())
        if len(teams) != 1:
            if raise_on_invalid_team:
                raise InvalidTeam()
            else:
                return None
        return self.get_team(teams[0])

    def get_all_teams(self) -> List[Team]:
        """
        Returns a list of all teams saved in the configuration file.

        Returns
        -------
        List[Team]
            The list of teams saved.
        """
        teams = list(self.get("teams").keys())
        teams_data: List[Team] = []
        for slug in teams:
            the_team_data = self.get_team(slug)
            if the_team_data:
                teams_data.append(the_team_data)

        return teams_data



from typing import List

# When updating this file, please also update the docs:
# https://docs.v7labs.com/docs/convert-1
supported_formats: List[str] = [
    "coco",
    "cvat",
    "dataloop",
    "darwin_1.0",
    "instance_mask",
    "pascalvoc",
    "semantic_mask",
    "semantic_mask_grey",
    "semantic_mask_index",
    "yolo",
    "nifti",
]



from importlib import import_module

from darwin.datatypes import ImportParser

from .importer import import_annotations  # noqa


class ImporterNotFoundError(ModuleNotFoundError):
    pass


def get_importer(format: str) -> ImportParser:
    try:
        module = import_module(f"darwin.importer.formats.{format}")
        return getattr(module, "parse_path")
    except ModuleNotFoundError:
        raise ImporterNotFoundError



import sys
from argparse import ArgumentParser, Namespace
from typing import Any, Optional, Tuple

import argcomplete


class Options:
    """
    Has functions to parse CLI options given by the user.
    """

    def __init__(self) -> None:

        self.parser: ArgumentParser = ArgumentParser(
            description="Command line tool to create/upload/download datasets on darwin."
        )

        subparsers = self.parser.add_subparsers(dest="command")
        subparsers.add_parser("help", help="Show this help message and exit.")

        # AUTHENTICATE
        subparsers.add_parser("authenticate", help="Authenticate the user. ")

        # SET COMPRESSION LEVEL
        parser_compression = subparsers.add_parser("compression", help="Set compression level.")
        parser_compression.add_argument(
            "compression_level",
            type=int,
            choices=range(0, 10),
            help="Compression level to use on uploaded data. 0 is no compression, 9 is the best.",
        )

        # SELECT TEAM
        parser_create = subparsers.add_parser("team", help="List or pick teams.")
        parser_create.add_argument("team_name", nargs="?", type=str, help="Team name to use.")
        parser_create.add_argument(
            "-c", "--current", action="store_true", required=False, help="Shows only the current team."
        )

        parser_convert = subparsers.add_parser("convert", help="Converts darwin json to other annotation formats.")
        parser_convert.add_argument("format", type=str, help="Annotation format to convert to.")

        parser_convert.add_argument("files", type=str, nargs="+", help="Annotation files (or folders) to convert.")

        parser_convert.add_argument("output_dir", type=str, help="Where to store output files.")

        # VALIDATE SCHEMA
        parser_validate_schema = subparsers.add_parser(
            "validate", help="Validate annotation files against Darwin schema"
        )
        parser_validate_schema.add_argument(
            "location",
            help="Location of file/folder to validate. Accepts single files or a folder to search *.json files",
        )
        parser_validate_schema.add_argument(
            "--pattern",
            action="store_true",
            help="'location' is a Folder + File glob style pattern to search (eg: ./*.json)",
        )

        parser_validate_schema.add_argument(
            "--silent", action="store_true", help="Flag to suppress all output except errors to console"
        )
        parser_validate_schema.add_argument("--output", help="name of file to write output json to")
        # DATASET
        dataset = subparsers.add_parser(
            "dataset", help="Dataset related functions.", description="Arguments to interact with datasets"
        )
        dataset_action = dataset.add_subparsers(dest="action")

        # Remote
        parser_remote = dataset_action.add_parser("remote", help="List remote datasets.")
        parser_remote.add_argument("-t", "--team", help="Specify team.")
        parser_remote.add_argument("-a", "--all", action="store_true", help="List datasets for all teams.")

        # Local
        parser_local = dataset_action.add_parser("local", help="List downloaded datasets.")
        parser_local.add_argument("-t", "--team", help="Specify team.")

        # Create
        parser_create = dataset_action.add_parser("create", help="Creates a new dataset on darwin.")
        parser_create.add_argument("dataset", type=str, help="Dataset name.")

        # Path
        parser_path = dataset_action.add_parser("path", help="Print local path to dataset.")
        parser_path.add_argument("dataset", type=str, help="Dataset name.")

        # Url
        parser_url = dataset_action.add_parser("url", help="Print url to dataset on darwin.")
        parser_url.add_argument("dataset", type=str, help="Dataset name.")

        # Push
        parser_push = dataset_action.add_parser("push", help="Upload data to an existing (remote) dataset.")
        parser_push.add_argument(
            "dataset",
            type=str,
            help="[Remote] Dataset name: to list all the existing dataset, run 'darwin dataset remote'.",
        )
        parser_push.add_argument("files", type=str, nargs="+", help="Files to upload.")
        parser_push.add_argument(
            "-e",
            "--exclude",
            type=str,
            nargs="+",
            default="",
            help="Excludes the files with the specified extension/s if a data folder is provided as data path.",
        )
        parser_push.add_argument(
            "-f",
            "--fps",
            default="native",
            help="Frames per second for video split (recommended: 1), use 'native' to use the videos intrinsic fps.",
        )
        parser_push.add_argument("--frames", action="store_true", help="Annotate a video as independent frames.")

        parser_push.add_argument(
            "--extract_views", action="store_true", help="Upload a volume with all 3 orthogonal views."
        )

        parser_push.add_argument("--path", type=str, default=None, help="Folder to upload the files into.")

        parser_push.add_argument("--verbose", action="store_true", help="Flag to show upload details.")

        parser_push.add_argument(
            "-p", "--preserve-folders", action="store_true", help="Preserve the local folder structure in the dataset."
        )

        # Remove
        parser_remove = dataset_action.add_parser("remove", help="Remove a remote or remote and local dataset.")
        parser_remove.add_argument("dataset", type=str, help="Remote dataset name to delete.")

        # Report
        parser_report = dataset_action.add_parser("report", help="Report about the annotators.")
        parser_report.add_argument("dataset", type=str, help="Remote dataset name to report on.")
        parser_report.add_argument(
            "-g", "--granularity", choices=["day", "week", "month", "total"], help="Granularity of the report."
        )
        parser_report.add_argument(
            "-r",
            "--pretty",
            action="store_true",
            default=False,
            help="Prints the results formatted in a rich table.",
        )
        # Export
        parser_export = dataset_action.add_parser("export", help="Export a version of a dataset.")
        parser_export.add_argument("dataset", type=str, help="Remote dataset name to export.")
        parser_export.add_argument("name", type=str, help="Name with with the version gets tagged.")
        parser_export.add_argument(
            "--class-ids",
            type=str,
            nargs="+",
            help=(
                "List of annotation class ids. If present, it will only include items that have"
                " annotations with a class whose id matches."
            ),
        )
        parser_export.add_argument(
            "--include-authorship",
            default=False,
            action="store_true",
            help="Each annotation contains annotator and reviewer authorship metadata.",
        )
        parser_export.add_argument(
            "--include-url-token",
            default=False,
            action="store_true",
            help="Each annotation file includes a url with an access token. "
            "Warning, anyone with the url can access the images, even without being a team member.",
        )
        parser_export.add_argument(
            "--version",
            default=None,
            type=str,
            choices=["1.0", "2.0"],
            help="When used for V2 dataset, allows to force generation of either Darwin JSON 1.0 (Legacy) or newer 2.0. "
            "Omit this option to get your team's default.",
        )

        # Releases
        parser_dataset_version = dataset_action.add_parser("releases", help="Available version of a dataset.")
        parser_dataset_version.add_argument("dataset", type=str, help="Remote dataset name to list.")

        # Pull
        parser_pull = dataset_action.add_parser("pull", help="Download a version of a dataset.")
        parser_pull.add_argument("dataset", type=str, help="Remote dataset name to download.")
        parser_pull.add_argument(
            "--only-annotations", action="store_true", help="Download only annotations and no corresponding images."
        )
        parser_pull.add_argument("--folders", action="store_true", help="Recreates image folders.")
        parser_pull.add_argument(
            "--video-frames", action="store_true", help="Pulls video frame images instead of video files."
        )
        parser_pull.add_argument(
            "--force-slots",
            action="store_true",
            help="Forces pull of all slots of items into deeper file structure ({prefix}/{item_name}/{slot_name}/{file_name}). "
            + "If your dataset includes items with multiple slots, or multiple source files per slot, this option becomes implicitly enabled.",
        )

        # Import
        parser_import = dataset_action.add_parser("import", help="Import data to an existing (remote) dataset.")
        parser_import.add_argument(
            "dataset",
            type=str,
            help="[Remote] Dataset name: to list all the existing dataset, run 'darwin dataset remote'.",
        )
        parser_import.add_argument("format", type=str, help="The format of the annotations to import.")

        parser_import.add_argument(
            "files",
            type=str,
            nargs="+",
            help="The location of the annotation files, or the folder where the annotation files are.",
        )
        parser_import.add_argument("--append", action="store_true", help="Append annotations instead of overwriting.")
        parser_import.add_argument(
            "--yes", action="store_true", help="Skips prompts for creating and adding classes to dataset."
        )
        parser_import.add_argument(
            "--delete-for-empty",
            action="store_true",
            help="Empty annotations will delete annotations from remote files.",
        )
        parser_import.add_argument(
            "--import-annotators",
            action="store_true",
            help="Import annotators metadata from the annotation files, where available",
        )
        parser_import.add_argument(
            "--import-reviewers",
            action="store_true",
            help="Import reviewers metadata from the annotation files, where available",
        )

        # Cpu limit for multiprocessing tasks
        def cpu_default_types(input: Any) -> Optional[int]:  # type: ignore
            try:
                return int(input)
            except TypeError:
                return None

        parser_import.add_argument(
            "--cpu-limit",
            "--cpu_limit",
            type=cpu_default_types,
            required=False,
            default=1,
            help="Limits amount of cores used on machine to process results, default to single core",
        )

        # Convert
        parser_convert = dataset_action.add_parser("convert", help="Converts darwin json to other annotation formats.")

        parser_convert.add_argument(
            "dataset",
            type=str,
            help="[Remote] Dataset name: to list all the existing dataset, run 'darwin dataset remote'.",
        )
        parser_convert.add_argument("format", type=str, help="Annotation format to convert to.")

        parser_convert.add_argument("-o", "--output_dir", type=str, help="Where to store output files.")

        # Split
        parser_split = dataset_action.add_parser(
            "split", help="Splits a local dataset following random and stratified split types."
        )
        parser_split.add_argument("dataset", type=str, help="Local dataset name to split.")
        parser_split.add_argument("-v", "--val-percentage", required=True, type=float, help="Validation percentage.")
        parser_split.add_argument("-t", "--test-percentage", required=True, type=float, help="Test percentage.")
        parser_split.add_argument("-s", "--seed", type=int, required=False, default=0, help="Split seed.")

        # List Files
        parser_files = dataset_action.add_parser("files", help="Lists file in a remote dataset.")
        parser_files.add_argument(
            "dataset",
            type=str,
            help="[Remote] Dataset name: to list all the existing dataset, run 'darwin dataset remote'.",
        )
        parser_files.add_argument("--only-filenames", action="store_true", help="Only prints out filenames.")
        parser_files.add_argument("--status", type=str, required=False, help="Comma separated list of statuses.")
        parser_files.add_argument(
            "--path",
            type=str,
            required=False,
            help="List only files under PATH. This is useful if your dataset has a directory structure.",
        )
        parser_files.add_argument(
            "--sort-by",
            type=str,
            required=False,
            help="Sort remotely fetched files by the given direction. Defaults to 'updated_at:desc'.",
        )

        # Set file status
        parser_file_status = dataset_action.add_parser("set-file-status", help="Sets the status of one or more files.")
        parser_file_status.add_argument(
            "dataset",
            type=str,
            help="[Remote] Dataset name: to list all the existing dataset, run 'darwin dataset remote'.",
        )
        parser_file_status.add_argument("status", type=str, help="Status to change to.")
        parser_file_status.add_argument("files", type=str, nargs="+", help="Files to change status.")

        # Delete files
        parser_delete_files = dataset_action.add_parser("delete-files", help="Delete one or more files remotely.")
        parser_delete_files.add_argument(
            "dataset",
            type=str,
            help="[Remote] Dataset name: to list all the existing dataset, run 'darwin dataset remote'.",
        )
        parser_delete_files.add_argument("files", type=str, nargs="+", help="Files to delete.")
        parser_delete_files.add_argument(
            "-y",
            "--yes",
            default=False,
            action="store_true",
            required=False,
            help="Confirmation flag to delete the file without prompting for manual input.",
        )

        # Add comments
        parser_comment = dataset_action.add_parser("comment", help="Comment image.")
        parser_comment.add_argument(
            "dataset",
            type=str,
            help="[Remote] Dataset name: to list all the existing dataset, run 'darwin dataset remote'. ",
        )
        parser_comment.add_argument("file", type=str, help="File to comment")
        parser_comment.add_argument("--text", type=str, help="Comment: list of words")
        parser_comment.add_argument("--x", required=False, type=float, default=1, help="X coordinate for comment box")
        parser_comment.add_argument("--y", required=False, type=float, default=1, help="Y coordinate for comment box")
        parser_comment.add_argument(
            "--w", "--width", required=False, type=float, default=1, help="Comment box width in pixels"
        )
        parser_comment.add_argument(
            "--h", "--height", required=False, type=float, default=1, help="Comment box height in pixels"
        )

        # Help
        dataset_action.add_parser("help", help="Show this help message and exit.")

        # VERSION
        subparsers.add_parser("version", help="Check current version of the repository. ")

        argcomplete.autocomplete(self.parser)

    def parse_args(self) -> Tuple[Namespace, ArgumentParser]:
        """
        Parses and validates the CLI options.

        Returns
        -------
        Tuple[Namespace, ArgumentParser]
            The tuple with the namespace and parser to use.
        """
        args = self.parser.parse_args()

        if not args.command:
            self.parser.print_help()
            sys.exit()

        return args, self.parser



import warnings
import zipfile
from collections import OrderedDict, defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Union

import orjson as json
from rich.console import Console

from darwin.utils import attempt_decode

console = Console()
try:
    import cc3d
    import nibabel as nib
except ImportError:
    import_fail_string = """
    You must install darwin-py with pip install darwin-py\[medical]
    in order to import with using nifti format
    """
    console.print(import_fail_string)
    exit()
import numpy as np
from jsonschema import validate
from upolygon import find_contours

import darwin.datatypes as dt
from darwin.importer.formats.nifti_schemas import nifti_import_schema
from darwin.version import __version__


def parse_path(path: Path) -> Optional[List[dt.AnnotationFile]]:
    """
    Parses the given ``nifti`` file and returns a ``List[dt.AnnotationFile]`` with the parsed
    information.

    Parameters
    ----------
    path : Path
        The ``Path`` to the ``nifti`` file.

    Returns
    -------
    Optional[List[dt.AnnotationFile]]
        Returns ``None`` if the given file is not in ``json`` format, or ``List[dt.AnnotationFile]``
        otherwise.
    """
    if not isinstance(path, Path):
        path = Path(path)
    if path.suffix != ".json":
        console.print("Skipping file: {} (not a json file)".format(path), style="bold yellow")
        return None
    data = attempt_decode(path)
    try:
        validate(data, schema=nifti_import_schema)
    except Exception as e:
        console.print("Skipping file: {} (invalid json file, see schema for details)".format(path), style="bold yellow")
        return None
    nifti_annotations = data.get("data")
    if nifti_annotations is None or nifti_annotations == []:
        console.print("Skipping file: {} (no data found)".format(path), style="bold yellow")
        return None
    annotation_files = []
    for nifti_annotation in nifti_annotations:
        annotation_file = _parse_nifti(
            Path(nifti_annotation["label"]),
            nifti_annotation["image"],
            path,
            class_map=nifti_annotation.get("class_map"),
            mode=nifti_annotation.get("mode", "image"),
        )
        annotation_files.append(annotation_file)
    return annotation_files


def _parse_nifti(nifti_path: Path, filename: Path, json_path: Path, class_map: Dict, mode: str) -> dt.AnnotationFile:
    img: np.ndarray = process_nifti(nib.load(nifti_path))

    shape = img.shape
    processed_class_map = process_class_map(class_map)
    video_annotations = []
    if mode == "instances":  # For each instance produce a video annotation
        for class_name, class_idxs in processed_class_map.items():
            if class_name == "background":
                continue
            class_img = np.isin(img, class_idxs).astype(np.uint8)
            cc_img, num_labels = cc3d.connected_components(class_img, return_N=True)
            for instance_id in range(1, num_labels):
                video_annotation = get_video_annotation(cc_img, class_idxs=[instance_id], class_name=class_name)
                if video_annotation:
                    video_annotations.append(video_annotation)
    elif mode == "image":  # For each frame and each class produce a single frame video annotation
        for i in range(shape[-1]):
            slice_mask = img[:, :, i].astype(np.uint8)
            for class_name, class_idxs in processed_class_map.items():
                frame_annotations = {}
                if class_name == "background":
                    continue
                class_mask = np.isin(slice_mask, class_idxs).astype(np.uint8).copy()
                polygon = mask_to_polygon(mask=class_mask, class_name=class_name)
                if polygon is None:
                    continue
                frame_annotations[i] = polygon
                video_annotation = dt.make_video_annotation(
                    frame_annotations,
                    keyframes={i: True, i + 1: True},
                    segments=[[i, i + 1]],
                    interpolated=False,
                    slot_names=[],
                )
                video_annotations.append(video_annotation)
    elif mode == "video":  # For each class produce a single video annotation
        for class_name, class_idxs in processed_class_map.items():
            if class_name == "background":
                continue
            video_annotation = get_video_annotation(img, class_idxs=class_idxs, class_name=class_name)
            if video_annotation is None:
                continue
            video_annotations.append(video_annotation)
    annotation_classes = set(
        [dt.AnnotationClass(class_name, "polygon", "polygon") for class_name in class_map.values()]
    )
    return dt.AnnotationFile(
        path=json_path,
        filename=str(filename),
        remote_path="/",
        annotation_classes=annotation_classes,
        annotations=video_annotations,
        slots=[dt.Slot(name=None, type="dicom", source_files=[{"url": None, "file_name": str(filename)}])],
    )


def get_video_annotation(volume: np.ndarray, class_name: str, class_idxs: List[int]) -> Optional[dt.VideoAnnotation]:
    frame_annotations = OrderedDict()
    for i in range(volume.shape[-1]):
        slice_mask = volume[:, :, i].astype(np.uint8)
        class_mask = np.isin(slice_mask, class_idxs).astype(np.uint8).copy()
        if class_mask.sum() == 0:
            continue

        polygon = mask_to_polygon(mask=class_mask, class_name=class_name)
        if polygon is None:
            continue
        frame_annotations[i] = polygon
    all_frame_ids = list(frame_annotations.keys())
    if not all_frame_ids:
        return None
    if len(all_frame_ids) == 1:
        segments = [[all_frame_ids[0], all_frame_ids[0] + 1]]
    elif len(all_frame_ids) > 1:
        segments = [[min(all_frame_ids), max(all_frame_ids)]]
    video_annotation = dt.make_video_annotation(
        frame_annotations,
        keyframes={f_id: True for f_id in all_frame_ids},
        segments=segments,
        interpolated=False,
        slot_names=[],
    )
    return video_annotation


def mask_to_polygon(mask: np.ndarray, class_name: str) -> Optional[dt.Annotation]:
    _labels, external_paths, _internal_paths = find_contours(mask)
    # annotations = []
    if len(external_paths) > 1:
        paths = []
        for external_path in external_paths:
            # skip paths with less than 2 points
            if len(external_path) // 2 <= 2:
                continue
            path = [{"x": y, "y": x} for x, y in zip(external_path[0::2], external_path[1::2])]
            paths.append(path)
        if len(paths) > 1:
            polygon = dt.make_complex_polygon(class_name, paths)
        elif len(paths) == 1:
            polygon = dt.make_polygon(
                class_name,
                point_path=paths[0],
            )
        else:
            return None
    elif len(external_paths) == 1:
        external_path = external_paths[0]
        if len(external_path) < 6:
            return None
        polygon = dt.make_polygon(
            class_name,
            point_path=[{"x": y, "y": x} for x, y in zip(external_path[0::2], external_path[1::2])],
        )
    else:
        return None
    return polygon


def process_class_map(class_map):
    """
    This function takes a class_map and returns a dictionary with the class names as keys and
    all the corresponding class indexes as values.
    """
    processed_class_map = defaultdict(list)
    for key, value in class_map.items():
        processed_class_map[value].append(int(key))
    return processed_class_map


def rectify_header_sform_qform(img_nii):
    """
    Look at the sform and qform of the nifti object and correct it if any
    incompatibilities with pixel dimensions

    Adapted from https://github.com/NifTK/NiftyNet/blob/v0.6.0/niftynet/io/misc_io.py

    Args:
        img_nii: nifti image object
    """
    d = img_nii.header["dim"][0]
    pixdim = np.asarray(img_nii.header.get_zooms())[:d]
    sform, qform = img_nii.get_sform(), img_nii.get_qform()
    norm_sform = affine_to_spacing(sform, r=d)
    norm_qform = affine_to_spacing(qform, r=d)
    sform_mismatch = not np.allclose(norm_sform, pixdim)
    qform_mismatch = not np.allclose(norm_qform, pixdim)

    if img_nii.header["sform_code"] != 0:
        if not sform_mismatch:
            return img_nii
        if not qform_mismatch:
            img_nii.set_sform(img_nii.get_qform())
            return img_nii
    if img_nii.header["qform_code"] != 0:
        if not qform_mismatch:
            return img_nii
        if not sform_mismatch:
            img_nii.set_qform(img_nii.get_sform())
            return img_nii

    norm = affine_to_spacing(img_nii.affine, r=d)
    warnings.warn(f"Modifying image pixdim from {pixdim} to {norm}")

    img_nii.header.set_zooms(norm)
    return img_nii


def affine_to_spacing(affine: np.ndarray, r: int = 3, dtype=float, suppress_zeros: bool = True) -> np.ndarray:
    """
    Copied over from monai.data.utils - https://docs.monai.io/en/stable/_modules/monai/data/utils.html

    Computing the current spacing from the affine matrix.

    Args:
        affine: a d x d affine matrix.
        r: indexing based on the spatial rank, spacing is computed from `affine[:r, :r]`.
        dtype: data type of the output.
        suppress_zeros: whether to surpress the zeros with ones.

    Returns:
        an `r` dimensional vector of spacing.
    """
    spacing = np.sqrt(np.sum(affine[:r, :r] * affine[:r, :r], axis=0))
    if suppress_zeros:
        spacing[spacing == 0] = 1.0
    return spacing


def correct_nifti_header_if_necessary(img_nii):
    """
    Check nifti object header's format, update the header if needed.
    In the updated image pixdim matches the affine.

    Args:
        img_nii: nifti image object
    """
    if img_nii.header.get("dim") is None:
        return img_nii  # not nifti?
    dim = img_nii.header["dim"][0]
    if dim >= 5:
        return img_nii  # do nothing for high-dimensional array
    # check that affine matches zooms
    pixdim = np.asarray(img_nii.header.get_zooms())[:dim]
    norm_affine = affine_to_spacing(img_nii.affine, r=dim)
    if np.allclose(pixdim, norm_affine):
        return img_nii
    if hasattr(img_nii, "get_sform"):
        return rectify_header_sform_qform(img_nii)
    return img_nii


def process_nifti(input_data: Union[Sequence[nib.nifti1.Nifti1Image], nib.nifti1.Nifti1Image]):
    """
    Function which takes in a single nifti path or a list of nifti paths
    and returns the pixel_array, affine and pixdim
    """
    if isinstance(input_data, nib.nifti1.Nifti1Image):
        img = correct_nifti_header_if_necessary(input_data)
        img = nib.funcs.as_closest_canonical(img)
        axcodes = nib.orientations.aff2axcodes(img.affine)
        # TODO: Future feature to pass custom ornt could go here.
        ornt = [[0.0, -1.0], [1.0, -1.0], [1.0, -1.0]]
        data_array = nib.orientations.apply_orientation(img.get_fdata(), ornt)
        return data_array



import unittest

import pytest
import responses
from pydantic import ValidationError

from darwin.future.core import backend as be
from darwin.future.core.client import Client
from darwin.future.data_objects.team import Team, TeamMember
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.fixtures import *


def test_get_team_returns_valid_team(base_client: Client, base_team_json: dict, base_team: Team) -> None:
    slug = "test-slug"
    endpoint = base_client.config.api_endpoint + f"teams/{slug}"
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, endpoint, json=base_team_json)

        team = be.get_team(base_client, slug)
        assert team == base_team


def test_get_team_fails_on_incorrect_input(base_client: Client, base_team: Team) -> None:
    slug = "test-slug"
    endpoint = base_client.config.api_endpoint + f"teams/{slug}"
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, endpoint, json={})

        with pytest.raises(ValidationError):
            team = be.get_team(base_client, slug)


def test_get_team_members_returns_valid_list(base_client: Client, base_team_member_json: dict) -> None:
    synthetic_list = [TeamMember.parse_obj(base_team_member_json), TeamMember.parse_obj(base_team_member_json)]
    endpoint = base_client.config.api_endpoint + "memberships"
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, endpoint, json=[base_team_member_json, base_team_member_json])

        members, errors = be.get_team_members(base_client)
        assert len(members) == 2
        assert len(errors) == 0
        assert members == synthetic_list


def test_get_team_members_fails_on_incorrect_input(base_client: Client, base_team_member_json: dict) -> None:
    endpoint = base_client.config.api_endpoint + "memberships"
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, endpoint, json=[base_team_member_json, {}])

        members, errors = be.get_team_members(base_client)
        assert len(members) == 1
        assert len(errors) == 1
        assert isinstance(errors[0], ValidationError)
        assert isinstance(members[0], TeamMember)



from typing import List, Optional

from pydantic import validator

from darwin.future.data_objects.dataset import DatasetList
from darwin.future.data_objects.team_member_role import TeamMemberRole
from darwin.future.data_objects.validators import parse_name
from darwin.future.pydantic_base import DefaultDarwin


class TeamMember(DefaultDarwin):
    """A class to manage all the information around a team member on the darwin platform
    Attributes
    ----------
    name: str
    """

    email: str
    first_name: str
    last_name: str
    role: TeamMemberRole
    team_id: int
    id: int
    user_id: int


TeamMemberList = List[TeamMember]


class Team(DefaultDarwin):
    """A class to manage all the information around a Team on the darwin platform, including validation

    Attributes
    ----------
    slug : str
    datasets: Optional[List[Dataset]] = None
        - a list of datasets linked to the team
    members: Optional[List[Release]] = None
        - a list of members linked to a team
    Methods
    ----------
    _slug_validator: validates and auto formats the slug variable
    """

    slug: str
    id: int
    datasets: Optional[DatasetList] = None
    members: Optional[List[TeamMember]] = None
    default_role: TeamMemberRole = TeamMemberRole.USER

    # Data Validation
    _slug_validator = validator("slug", allow_reuse=True)(parse_name)


TeamList = List[Team]



from dataclasses import dataclass
from typing import Any, Dict, List, Optional

import deprecation
from pydantic import BaseModel

from darwin.path_utils import construct_full_path
from darwin.version import __version__


@dataclass(frozen=True, eq=True)
class DatasetItem(BaseModel):
    """
    DatasetItem represents files that can be images or videos which belong to a dataset.
    """

    #: The id of this ``DatasetItem``.
    id: int

    #: The filename of this ``DatasetItem``.
    filename: str

    #: The status of this ``DatasetItem``. It can be ``"archived"``, ``"error"``, ``"uploading"``,
    #: ``"processing"``, ``"new"``, ``"annotate"``, ``"review"`` or ``"complete"``.
    status: str

    #: Whether or not this item was soft deleted.
    archived: bool

    #: The size of this ``DatasetItem``\'s file in bytes.
    filesize: int

    #: The id of the ``Dataset`` this ``DatasetItem`` belongs to.
    dataset_id: int

    #: The slugified name of the ``Dataset`` this ``DatasetItem`` belongs to.
    dataset_slug: str

    #: The sequential value of this ``DatasetItem`` in relation to the ``Dataset`` it belongs to.
    #: This allows us to know which items were added first and is used mostly for sorting purposes.
    seq: int

    #: The id of this ``DatasetItem``'s workflow. A ``None`` value means this ``DatasetItem`` is
    #: new and was never worked on, or was reset to the new state.
    current_workflow_id: Optional[int]

    #: The darwin path to this ``DatasetItem``.
    path: str

    #: The names of each slot in the item, most items have a single slot corresponding to the file itself.
    #: only used for v2 dataset items
    slots: List[Any]

    #: Metadata of this ``DatasetItem``'s workflow. A ``None`` value means this ``DatasetItem`` is
    #: new and was never worked on, or was reset to the new state.
    current_workflow: Optional[Dict[str, Any]]

    @property
    def full_path(self) -> str:
        """
        The full POSIX relative path of this ``DatasetItem``.
        """
        return construct_full_path(self.path, self.filename)

    @classmethod
    def parse(cls, raw: Dict[str, Any]) -> "DatasetItem":
        """
        Parses the given dictionary into a ``DatasetItem``.

        Parameters
        ----------
        raw : Dict[str, Any]
            The dictionary to parse.

        Returns
        -------
        DatasetItem
            A dataset item with the parsed information.

        Raises
        ------
        ValidationError
            If any of the keys from the given dictionary do not have the correct format or are
            missing.
        """
        if "slots" in raw:
            data = {
                "id": raw["id"],
                "filename": raw["name"],
                "path": raw["path"],
                "status": raw["status"],
                "archived": raw["archived"],
                "filesize": sum(file.get("size_bytes", 0) for file in raw["slots"]),
                "dataset_id": raw["dataset_id"],
                "dataset_slug": "n/a",
                "seq": None,
                "current_workflow_id": None,
                "current_workflow": None,
                "slots": raw["slots"],
                "current_workflow": None,
            }
        else:
            data = {
                "id": raw["id"],
                "filename": raw["filename"],
                "status": raw["status"],
                "archived": raw["archived"],
                "filesize": raw["file_size"],
                "dataset_id": raw["dataset_id"],
                "dataset_slug": "n/a",
                "seq": raw["seq"],
                "current_workflow_id": raw.get("current_workflow_id"),
                "current_workflow": raw.get("current_workflow"),
                "path": raw["path"],
                "slots": [],
            }
        return DatasetItem(**data)


@deprecation.deprecated(
    deprecated_in="0.7.5",
    removed_in="0.8.0",
    current_version=__version__,
    details="Use the ``DatasetItem.parse`` instead.",
)
def parse_dataset_item(raw: Dict[str, Any]) -> DatasetItem:
    """
    Parses the given dictionary into a ``DatasetItem``. Performs no validations.

    Parameters
    ----------
    raw : Dict[str, Any]
        The dictionary to parse.

    Returns
    -------
    DatasetItem
        A dataset item with the parsed information.
    """
    return DatasetItem(
        raw["id"],
        raw["filename"],
        raw["status"],
        raw["archived"],
        raw["file_size"],
        raw["dataset_id"],
        "n/a",
        raw["seq"],
        raw.get("current_workflow_id"),
        raw["path"],
        [],
        raw.get("current_workflow"),
    )



import random
from typing import Any, Dict, Optional, Tuple, Union

import torch
import torchvision.transforms as transforms
import torchvision.transforms.functional as F
from PIL import Image as PILImage

from darwin.torch.utils import convert_segmentation_to_mask, flatten_masks_by_category


TargetKey = Union["boxes", "labels", "mask", "masks", "image_id", "area", "iscrowd"]
TargetType = Dict[TargetKey, torch.Tensor]


class Compose(transforms.Compose):
    """
    Composes a sequence of Transformations.
    """

    def __call__(self, image: PILImage.Image, target: Optional[TargetType] = None):
        if target is None:
            return super(Compose, self).__call__(image)
        for transform in self.transforms:
            image, target = transform(image, target)
        return image, target


class RandomHorizontalFlip(transforms.RandomHorizontalFlip):
    """
    Allows for horizontal flipping of an image, randomly.
    """

    def forward(
        self, image: torch.Tensor, target: Optional[TargetType] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]:
        """
        May or may not horizontally flip an image depending on a random factor.

        Parameters
        ----------
        image : torch.Tensor
            Image ``Tensor`` to flip.
        target : Optional[TargetType] = None
            The target.

        Returns
        -------
        Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]
            Will return a single image ``Tensor`` if the flip did not happen, or a tuple of the
            image tensor and the target type if the flip did happen.

        """
        if random.random() < self.p:
            image = F.hflip(image)
            if target is None:
                return image

            if "boxes" in target:
                bbox = target["boxes"]
                bbox[:, [0, 2]] = image.size[0] - bbox[:, [2, 0]]
                target["boxes"] = bbox
            for k in ["mask", "masks"]:
                if k in target:
                    target[k] = target[k].flip(-1)
            return image, target

        if target is None:
            return image

        return image, target


class RandomVerticalFlip(transforms.RandomVerticalFlip):
    """
    Allows for vertical flipping of an image, randomly.
    """

    def forward(
        self, image: torch.Tensor, target: Optional[TargetType] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]:
        """
        May or may not vertically flip an image depending on a random factor.

        Parameters
        ----------
        image : torch.Tensor
            Image ``Tensor`` to flip.
        target : Optional[TargetType] = None
            The target.

        Returns
        -------
        Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]
            Will return a single image ``Tensor`` if the flip did not happen, or a tuple of the
            image tensor and the target type if the flip did happen.

        """
        if random.random() < self.p:
            image = F.vflip(image)
            if target is None:
                return image

            if "boxes" in target:
                bbox = target["boxes"]
                bbox[:, [1, 3]] = image.size[1] - bbox[:, [1, 3]]
                target["boxes"] = bbox
            for k in ["mask", "masks"]:
                if k in target:
                    target[k] = target[k].flip(-2)
            return image, target

        if target is None:
            return image
        return image, target


class ColorJitter(transforms.ColorJitter):
    """
    Jitters the colors of the given transformation.
    """

    def __call__(
        self, image: PILImage.Image, target: Optional[TargetType] = None
    ) -> Union[PILImage.Image, Tuple[PILImage.Image, TargetType]]:
        transform = self.get_params(self.brightness, self.contrast, self.saturation, self.hue)
        image = transform(image)
        if target is None:
            return image
        return image, target


class ToTensor(transforms.ToTensor):
    """
    Converts given ``PILImage`` to a ``Tensor``.
    """

    def __call__(
        self, image: PILImage.Image, target: Optional[TargetType] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]:
        image_tensor: torch.Tensor = F.to_tensor(image)
        if target is None:
            return image_tensor
        return image_tensor, target


class ToPILImage(transforms.ToPILImage):
    """
    Converts given ``Tensor`` to a ``PILImage``.
    """

    def __call__(
        self, image: torch.Tensor, target: Optional[TargetType] = None
    ) -> Union[PILImage.Image, Tuple[PILImage.Image, TargetType]]:
        pil_image: PILImage.Image = F.to_pil_image(image)
        if target is None:
            return pil_image
        return pil_image, target


class Normalize(transforms.Normalize):
    """
    Normalizes the given ``Tensor``.
    """

    def __call__(
        self, tensor: torch.Tensor, target: Optional[TargetType] = None
    ) -> Union[torch.Tensor, Tuple[torch.Tensor, TargetType]]:
        tensor = F.normalize(tensor, self.mean, self.std, self.inplace)

        if target is None:
            return tensor
        return tensor, target


class ConvertPolygonsToInstanceMasks(object):
    """
    Converts given polygon to an ``InstanceMask``.
    """

    def __call__(self, image: PILImage.Image, target: TargetType) -> Tuple[PILImage.Image, TargetType]:
        w, h = image.size

        image_id = target["image_id"]
        image_id = torch.tensor([image_id])

        annotations = target.pop("annotations")

        annotations = [obj for obj in annotations if obj.get("iscrowd", 0) == 0]

        boxes = [obj["bbox"] for obj in annotations]
        # guard against no boxes via resizing
        boxes = torch.as_tensor(boxes, dtype=torch.float32).reshape(-1, 4)
        boxes[:, 2:] += boxes[:, :2]
        boxes[:, 0::2].clamp_(min=0, max=w)
        boxes[:, 1::2].clamp_(min=0, max=h)

        classes = [obj["category_id"] for obj in annotations]
        classes = torch.tensor(classes, dtype=torch.int64)

        segmentations = [obj["segmentation"] for obj in annotations]
        masks = convert_segmentation_to_mask(segmentations, h, w)

        keypoints = None
        if annotations and "keypoints" in annotations[0]:
            keypoints = [obj["keypoints"] for obj in annotations]
            keypoints = torch.as_tensor(keypoints, dtype=torch.float32)
            num_keypoints = keypoints.shape[0]
            if num_keypoints:
                keypoints = keypoints.view(num_keypoints, -1, 3)

        keep = (boxes[:, 3] > boxes[:, 1]) & (boxes[:, 2] > boxes[:, 0])
        boxes = boxes[keep]
        classes = classes[keep]
        masks = masks[keep]
        if keypoints is not None:
            keypoints = keypoints[keep]

        target["boxes"] = boxes
        target["labels"] = classes
        target["masks"] = masks
        target["image_id"] = image_id
        if keypoints is not None:
            target["keypoints"] = keypoints

        # conversion to coco api
        area = torch.tensor([obj["area"] for obj in annotations])
        iscrowd = torch.tensor([obj.get("iscrowd", 0) for obj in annotations])
        target["area"] = area
        target["iscrowd"] = iscrowd

        return image, target


class ConvertPolygonsToSemanticMask(object):
    """
    Converts given polygon to an ``SemanticMask``.
    """

    def __call__(self, image: PILImage.Image, target: TargetType) -> Tuple[PILImage.Image, TargetType]:
        w, h = image.size
        image_id = target["image_id"]
        image_id = torch.tensor([image_id])

        annotations = target.pop("annotations")
        segmentations = [obj["segmentation"] for obj in annotations]
        cats = [obj["category_id"] for obj in annotations]
        if segmentations:
            masks = convert_segmentation_to_mask(segmentations, h, w)
            # merge all instance masks into a single segmentation map
            # with its corresponding categories
            mask = flatten_masks_by_category(masks, cats)

        else:
            mask = torch.zeros((h, w), dtype=torch.uint8)

        target["mask"] = mask
        target["image_id"] = image_id
        return image, target


class ConvertPolygonToMask(object):
    """
    Converts given polygon to a ``Mask``.
    """

    def __call__(self, image: PILImage.Image, annotation: Dict[str, Any]) -> Tuple[PILImage.Image, PILImage.Image]:
        w, h = image.size
        segmentations = [obj["segmentation"] for obj in annotation]
        cats = [obj["category_id"] for obj in annotation]
        if segmentations:
            masks = convert_segmentation_to_mask(segmentations, h, w)
            # merge all instance masks into a single segmentation map
            # with its corresponding categories
            target = flatten_masks_by_category(masks, cats)

        else:
            target = torch.zeros((h, w), dtype=torch.uint8)
        target = PILImage.fromarray(target.numpy())
        return image, target



import json
from typing import Any

import numpy as np


class NumpyEncoder(json.JSONEncoder):
    """
    Holds auxiliary functions to bridge numpy functionality with Python primitive types which are
    JSON friendly.
    """

    def default(self, obj: Any) -> Any:
        """
        Converts the given numpy object into a Python's primitive type.

        Parameters
        ----------
        obj : Any
            The object to convert.

        Returns
        -------
        Any
            The converted object.
        """
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyEncoder, self).default(obj)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['darwin',
 'darwin.dataset',
 'darwin.exporter',
 'darwin.exporter.formats',
 'darwin.future',
 'darwin.future.core',
 'darwin.future.core.types',
 'darwin.future.core.utils',
 'darwin.future.data_objects',
 'darwin.future.exceptions',
 'darwin.future.meta',
 'darwin.future.meta.types',
 'darwin.future.tests',
 'darwin.future.tests.core',
 'darwin.future.tests.data_objects',
 'darwin.future.tests.meta',
 'darwin.importer',
 'darwin.importer.formats',
 'darwin.torch',
 'darwin.utils',
 'darwin.version']

package_data = \
{'': ['*']}

install_requires = \
['argcomplete>=2.0.0,<3.0.0',
 'deprecation>=2.1.0,<3.0.0',
 'humanize>=4.4.0,<5.0.0',
 'jsonschema>=4.0.0',
 'mpire>=2.7.0,<3.0.0',
 'numpy',
 'orjson>=3.8.5,<4.0.0',
 'pillow>=9.4.0,<10.0.0',
 'pydantic>=1.10.4,<2.0.0',
 'pyyaml>=5.1',
 'requests>=2.28.1,<3.0.0',
 'rich>=13.0.1,<14.0.0',
 'toml>=0.10.2,<0.11.0',
 'tqdm>=4.64.1,<5.0.0',
 'types-pyyaml>=6.0.12.9,<7.0.0.0',
 'types-requests>=2.28.11.8,<3.0.0.0',
 'upolygon==0.1.10']

extras_require = \
{':(python_full_version >= "3.8.1" and python_version < "3.11") and (extra == "dev" or extra == "test")': ['flake8-pyproject>=1.2.2,<2.0.0'],
 ':(python_full_version >= "3.8.1" and python_version < "3.11") and (extra == "dev")': ['flake8>=6.0.0,<7.0.0'],
 ':(python_full_version >= "3.8.1" and python_version < "3.11") and (extra == "ml")': ['scikit-learn>=1.2.0,<2.0.0'],
 ':(python_full_version >= "3.8.1") and (extra == "medical")': ['nibabel>=5.0.0,<6.0.0'],
 ':extra == "dev"': ['black>=22.12.0,<23.0.0',
                     'isort>=5.11.4,<6.0.0',
                     'mypy>=0.991,<0.992',
                     'debugpy>=1.6.5,<2.0.0'],
 ':extra == "dev" or extra == "test"': ['responses>=0.22.0,<0.23.0',
                                        'pytest>=7.2.1,<8.0.0',
                                        'pytest-describe>=2.0.1,<3.0.0'],
 ':extra == "medical"': ['connected-components-3d>=3.10.3,<4.0.0'],
 ':extra == "ml"': ['torch>=1.13.1,<2.0.0', 'torchvision>=0.14.1,<0.15.0'],
 ':python_version < "3.7"': ['dataclasses>=0.8,<0.9']}

entry_points = \
{'console_scripts': ['darwin = darwin.cli:main']}

setup_kwargs = {
    'name': 'darwin-py',
    'version': '0.8.23',
    'description': 'Library and command line interface for darwin.v7labs.com',
    'long_description': '# V7 Darwin Python SDK\n\nOfficial library to manage datasets along with\n[V7 Darwin annotation platform](https://darwin.v7labs.com).\n\nDarwin-py can both be used from the [command line](#usage-as-a-command-line-interface-cli) and as a [python library](#usage-as-a-python-library).\n\nMain functions are (but not limited to):\n\n- Client authentication\n- Listing local and remote datasets\n- Create/remove datasets\n- Upload/download data to/from remote datasets\n- Direct integration with PyTorch dataloaders\n\nSupport tested for python 3.8.\n\n## Installation\n\n```\npip install darwin-py\n```\n\nYou can now type `darwin` in your terminal and access the command line interface.\n\nIf you wish to use the PyTorch bindings, then you can use the `ml` flag to install all the additional requirements\n\n```\npip install darwin-py[ml]\n```\n\nTo run test, first install the `test` extra package\n\n```\npip install darwin-py[test]\n```\n\n---\n\n## Usage as a Command Line Interface (CLI)\n\nOnce installed, `darwin` is accessible as a command line tool.\nA useful way to navigate the CLI usage is through the help command `-h/--help` which will\nprovide additional information for each command available.\n\n### Client Authentication\n\nTo perform remote operations on Darwin you first need to authenticate.\nThis requires a [team-specific API-key](https://darwin.v7labs.com/?settings=api-keys).\nIf you do not already have a Darwin account, you can [contact us](https://www.v7labs.com/contact) and we can set one up for you.\n\nTo start the authentication process:\n\n```\n$ darwin authenticate\nAPI key:\nMake example-team the default team? [y/N] y\nDatasets directory [~/.darwin/datasets]:\nAuthentication succeeded.\n```\n\nYou will be then prompted to enter your API-key, whether you want to set the corresponding team as\ndefault and finally the desired location on the local file system for the datasets of that team.\nThis process will create a configuration file at `~/.darwin/config.yaml`.\nThis file will be updated with future authentications for different teams.\n\n### Listing local and remote datasets\n\nLists a summary of local existing datasets\n\n```\n$ darwin dataset local\nNAME            IMAGES     SYNC_DATE         SIZE\nmydataset       112025     yesterday     159.2 GB\n```\n\nLists a summary of remote datasets accessible by the current user.\n\n```\n$ darwin dataset remote\nNAME                       IMAGES     PROGRESS\nexample-team/mydataset     112025        73.0%\n```\n\n### Create/remove a dataset\n\nTo create an empty dataset remotely:\n\n```\n$ darwin dataset create test\nDataset \'test\' (example-team/test) has been created.\nAccess at https://darwin.v7labs.com/datasets/579\n```\n\nThe dataset will be created in the team you\'re authenticated for.\n\nTo delete the project on the server:\n\n```\n$ darwin dataset remove test\nAbout to delete example-team/test on darwin.\nDo you want to continue? [y/N] y\n```\n\n### Upload/download data to/from a remote dataset\n\nUploads data to an existing remote project.\nIt takes the dataset name and a single image (or directory) with images/videos to upload as\nparameters.\n\nThe `-e/--exclude` argument allows to indicate file extension/s to be ignored from the data_dir.\ne.g.: `-e .jpg`\n\nFor videos, the frame rate extraction rate can be specified by adding `--fps <frame_rate>`\n\nSupported extensions:\n\n- Video files: [`.mp4`, `.bpm`, `.mov` formats].\n- Image files [`.jpg`, `.jpeg`, `.png` formats].\n\n```\n$ darwin dataset push test /path/to/folder/with/images\n100%|████████████████████████| 2/2 [00:01<00:00,  1.27it/s]\n```\n\nBefore a dataset can be downloaded, a release needs to be generated:\n\n```\n$ darwin dataset export test 0.1\nDataset test successfully exported to example-team/test:0.1\n```\n\nThis version is immutable, if new images / annotations have been added you will have to create a new release to included them.\n\nTo list all available releases\n\n```\n$ darwin dataset releases test\nNAME                           IMAGES     CLASSES                   EXPORT_DATE\nexample-team/test:0.1               4           0     2019-12-07 11:37:35+00:00\n```\n\nAnd to finally download a release.\n\n```\n$ darwin dataset pull test:0.1\nDataset example-team/test:0.1 downloaded at /directory/choosen/at/authentication/time.\n```\n\n---\n\n## Usage as a Python library\n\nThe framework is designed to be usable as a standalone python library.\nUsage can be inferred from looking at the operations performed in `darwin/cli_functions.py`.\nA minimal example to download a dataset is provided below and a more extensive one can be found in\n[darwin_demo.py]([./darwin_demo.py](https://github.com/v7labs/darwin-py/blob/master/darwin_demo.py).\n\n```python\nfrom darwin.client import Client\n\nclient = Client.local() # use the configuration in ~/.darwin/config.yaml\ndataset = client.get_remote_dataset("example-team/test")\ndataset.pull() # downloads annotations and images for the latest exported version\n```\n\nFollow [this guide](https://docs.v7labs.com/docs/loading-a-dataset-in-python) for how to integrate darwin datasets directly in PyTorch.\n',
    'author': 'V7',
    'author_email': 'info@v7labs.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://docs.v7labs.com/reference/getting-started-2',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<3.12',
}


setup(**setup_kwargs)



import sys
from pathlib import Path

try:
    import torchvision  # noqa

    # Here we remove `darwin` directory from `sys.path` to force the importer
    # to import the library `torch`, rather than the internal package.
    # This hack resolves this naming conflict for Sphinx.
    for path in sys.path:
        path_str = str(Path("darwin-py") / "darwin")
        if path.endswith(path_str):
            sys.path.remove(path)

    import torch  # noqa
except ImportError:
    raise ImportError(
        f"darwin.torch requires pytorch and torchvision. Install it using: pip install torch torchvision"
    ) from None

from .dataset import get_dataset  # noqa



import unittest
from pathlib import Path

import pytest
import responses
from pydantic import ValidationError
from requests import HTTPError

from darwin.future.core.client import Client, DarwinConfig, TeamsConfig
from darwin.future.exceptions.base import DarwinException
from darwin.future.exceptions.client import NotFound, Unauthorized
from darwin.future.tests.core.fixtures import *
from darwin.future.tests.fixtures import *


def test_create_config(base_config: DarwinConfig) -> None:
    assert base_config.api_key == "test_key"
    assert base_config.base_url == "http://test_url.com/"
    assert base_config.default_team == "default-team"


def test_config_base_url(base_config: DarwinConfig) -> None:
    assert base_config.base_url == "http://test_url.com/"

    # Test that the base_url validates after being created
    base_config.base_url = "https://test_url.com"
    assert base_config.base_url == "https://test_url.com/"

    # Test that the base_url fails validation on invalid url strings
    with pytest.raises(ValidationError):
        base_config.base_url = "test_url.com"
        base_config.base_url = "ftp://test_url.com"
        base_config.base_url = ""


@pytest.mark.parametrize("base_url", ["test_url.com", "ftp://test_url.com", ""])
def test_invalid_config_url_validation(base_url: str) -> None:
    with pytest.raises(ValidationError):
        config = DarwinConfig(
            api_key="test_key",
            api_endpoint="http://test_url.com/api/",
            base_url=base_url,
            default_team="default-team",
            teams={},
        )


def test_client(base_client: Client) -> None:
    assert base_client.config.api_key == "test_key"
    assert base_client.config.base_url == "http://test_url.com/"
    assert base_client.config.default_team == "default-team"

    assert base_client.session is not None
    assert base_client.headers is not None
    assert base_client.headers == {"Content-Type": "application/json", "Authorization": "ApiKey test_key"}

    # Test client functionality works
    endpoint = base_client.config.api_endpoint + "test_endpoint"
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, endpoint, json={"test": "test"}, status=200)
        rsps.add(responses.PUT, endpoint, json={"test": "test"}, status=200)
        rsps.add(responses.POST, endpoint, json={"test": "test"}, status=200)
        rsps.add(responses.DELETE, endpoint, json={"test": "test"}, status=200)
        rsps.add(responses.PATCH, endpoint, json={"test": "test"}, status=200)

        # Test get
        response = base_client.get("test_endpoint")
        assert response == {"test": "test"}

        # Test put
        response = base_client.put("test_endpoint", {"test": "test"})
        assert response == {"test": "test"}

        # Test post
        response = base_client.post("test_endpoint", {"test": "test"})
        assert response == {"test": "test"}

        # Test delete
        response = base_client.delete("test_endpoint")
        assert response == {"test": "test"}

        # Test patch
        response = base_client.patch("test_endpoint", {"test": "test"})
        assert response == {"test": "test"}


@pytest.mark.parametrize(
    "status_code, exception",
    [(401, Unauthorized), (404, NotFound)],
)
def test_client_raises_darwin(status_code: int, exception: DarwinException, base_client: Client) -> None:
    endpoint = base_client.config.api_endpoint + "test_endpoint"
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(exception):  # type: ignore
            base_client.get("test_endpoint")
        rsps.reset()
        rsps.add(responses.PUT, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(exception):  # type: ignore
            base_client.put("test_endpoint", {"test": "test"})
        rsps.reset()
        rsps.add(responses.POST, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(exception):  # type: ignore
            base_client.post("test_endpoint", {"test": "test"})
        rsps.reset()
        rsps.add(responses.DELETE, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(exception):  # type: ignore
            base_client.delete("test_endpoint")
        rsps.reset()
        rsps.add(responses.PATCH, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(exception):  # type: ignore
            base_client.patch("test_endpoint", {"test": "test"})


def test_client_raises_generic(base_client: Client) -> None:
    endpoint = base_client.config.api_endpoint + "test_endpoint"
    status_code = 499
    with responses.RequestsMock() as rsps:
        rsps.add(responses.GET, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(HTTPError):
            base_client.get("test_endpoint")
        rsps.reset()
        rsps.add(responses.PUT, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(HTTPError):
            base_client.put("test_endpoint", {"test": "test"})
        rsps.reset()
        rsps.add(responses.POST, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(HTTPError):
            base_client.post("test_endpoint", {"test": "test"})
        rsps.reset()
        rsps.add(responses.DELETE, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(HTTPError):
            base_client.delete("test_endpoint")
        rsps.reset()
        rsps.add(responses.PATCH, endpoint, json={"test": "test"}, status=status_code)
        with pytest.raises(HTTPError):
            base_client.patch("test_endpoint", {"test": "test"})



##################################
#       import_file.json         #
##################################

attributes = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": {"id": {"type": "integer"}, "groupId": {"type": "integer"}},
        "required": ["id", "groupId"],
    },
}

bbox = {
    "$id": "https://darwin.v7labs.com/schemas/supperannotate/bounding_box",
    "description": "Schema of a Bounding Box",
    "title": "Bounding Box",
    "default": {
        "type": "bbox",
        "points": {
            "x1": 1223.1,
            "x2": 1420.2,
            "y1": 607.3,
            "y2": 1440,
        },
        "classId": 1,
        "attributes": [],
    },
    "examples": [
        {
            "type": "bbox",
            "points": {
                "x1": 587.5,
                "x2": 1420.2,
                "y1": 607.3,
                "y2": 1440,
            },
            "classId": 1,
            "attributes": [{"id": 1, "groupId": 2}],
        }
    ],
    "type": "object",
    "properties": {
        "attributes": attributes,
        "classId": {"type": "integer"},
        "type": {"enum": ["bbox"]},
        "points": {
            "type": "object",
            "properties": {
                "x1": {
                    "type": "number",
                },
                "x2": {
                    "type": "number",
                },
                "y1": {
                    "type": "number",
                },
                "y2": {
                    "type": "number",
                },
            },
            "required": ["x1", "x2", "y1", "y2"],
        },
    },
    "required": ["points", "type", "classId", "attributes"],
}

polygon = {
    "$id": "https://darwin.v7labs.com/schemas/supperannotate/polygon",
    "description": "Schema of a Polygon",
    "title": "Polygon",
    "default": {"type": "polygon", "points": [1, 2, 3, 4], "classId": 1},
    "examples": [
        {"type": "polygon", "points": [1, 2, 3, 4], "classId": 1, "attributes": [{"id": 1, "groupId": 2}]},
        {"type": "polygon", "points": [], "classId": 1, "attributes": []},
    ],
    "type": "object",
    "properties": {
        "attributes": attributes,
        "classId": {"type": "integer"},
        "points": {"type": "array", "items": {"type": "number"}},
        "type": {"enum": ["polygon"]},
    },
    "required": ["points", "type", "classId", "attributes"],
}

polyline = {
    "$id": "https://darwin.v7labs.com/schemas/supperannotate/polyline",
    "description": "Schema of a Polyline",
    "title": "Polyline",
    "default": {"type": "polyline", "points": [1, 2, 3, 4], "classId": 1},
    "examples": [
        {"type": "polyline", "points": [1, 2, 3, 4], "classId": 1, "attributes": [{"id": 1, "groupId": 2}]},
        {"type": "polyline", "points": [], "classId": 1, "attributes": []},
    ],
    "type": "object",
    "properties": {
        "attributes": attributes,
        "classId": {"type": "integer"},
        "points": {"type": "array", "items": {"type": "number"}},
        "type": {"enum": ["polyline"]},
    },
    "required": ["points", "type", "classId", "attributes"],
}

cuboid = {
    "$id": "https://darwin.v7labs.com/schemas/supperannotate/cuboid",
    "description": "Schema of a Cuboid",
    "title": "Cuboid",
    "default": {
        "type": "cuboid",
        "points": {
            "f1": {"x": 1223.1, "y": 587.5},
            "f2": {"x": 1540.3, "y": 1420.2},
            "r1": {"x": 1286.2, "y": 607.3},
            "r2": {"x": 1603.4, "y": 1440},
        },
        "classId": 1,
        "attributes": [{"id": 1, "groupId": 2}],
    },
    "examples": [
        {
            "type": "cuboid",
            "points": {
                "f1": {"x": 1223.1, "y": 587.5},
                "f2": {"x": 1540.3, "y": 1420.2},
                "r1": {"x": 1286.2, "y": 607.3},
                "r2": {"x": 1603.4, "y": 1440},
            },
            "classId": 1,
            "attributes": [],
        }
    ],
    "type": "object",
    "properties": {
        "attributes": attributes,
        "classId": {"type": "integer"},
        "type": {"enum": ["cuboid"]},
        "points": {
            "type": "object",
            "properties": {
                "f1": {
                    "type": "object",
                    "properties": {"x": {"type": "number"}, "y": {"type": "number"}},
                    "required": ["x", "y"],
                },
                "f2": {
                    "type": "object",
                    "properties": {"x": {"type": "number"}, "y": {"type": "number"}},
                    "required": ["x", "y"],
                },
                "r1": {
                    "type": "object",
                    "properties": {"x": {"type": "number"}, "y": {"type": "number"}},
                    "required": ["x", "y"],
                },
                "r2": {
                    "type": "object",
                    "properties": {"x": {"type": "number"}, "y": {"type": "number"}},
                    "required": ["x", "y"],
                },
            },
            "required": ["f1", "f2", "r1", "r2"],
        },
    },
    "required": ["points", "type", "classId", "attributes"],
}

ellipse = {
    "$id": "https://darwin.v7labs.com/schemas/supperannotate/ellipse",
    "description": "Schema of an Ellipse",
    "title": "Ellipse",
    "default": {
        "type": "ellipse",
        "cx": 377.46,
        "cy": 806.18,
        "rx": 316.36,
        "ry": 134.18,
        "angle": 0,
        "classId": 1,
        "attributes": [],
    },
    "examples": [
        {
            "type": "ellipse",
            "cx": 377.46,
            "cy": 806.18,
            "rx": 316.36,
            "ry": 134.18,
            "angle": 14.66,
            "classId": 1,
            "attributes": [{"id": 1, "groupId": 2}],
        }
    ],
    "type": "object",
    "properties": {
        "attributes": attributes,
        "classId": {"type": "integer"},
        "cx": {"type": "number"},
        "cy": {"type": "number"},
        "rx": {"type": "number"},
        "ry": {"type": "number"},
        "angle": {"type": "number"},
        "type": {"enum": ["ellipse"]},
    },
    "required": ["cx", "cy", "rx", "ry", "angle", "type", "classId", "attributes"],
}

point = {
    "$id": "https://darwin.v7labs.com/schemas/supperannotate/point",
    "description": "Schema of a Point",
    "title": "Point",
    "default": {"type": "point", "x": 1.2, "y": 2.5, "classId": 1, "attributes": []},
    "examples": [
        {"type": "point", "x": 1.2, "y": 2.5, "classId": 1, "attributes": []},
        {"type": "point", "x": 0, "y": 1, "classId": 2, "attributes": [{"id": 1, "groupId": 2}]},
    ],
    "type": "object",
    "properties": {
        "attributes": attributes,
        "classId": {"type": "integer"},
        "x": {"type": "number"},
        "y": {"type": "number"},
        "type": {"enum": ["point"]},
    },
    "required": ["x", "y", "type", "classId", "attributes"],
}


superannotate_export = {
    "type": "object",
    "required": ["instances", "metadata", "tags"],
    "properties": {
        "instances": {
            "type": "array",
            "items": {"oneOf": [point, ellipse, cuboid, polygon, bbox, polyline]},
        },
        "metadata": {"type": "object", "required": ["name"], "properties": {"name": {"type": "string"}}},
        "tags": {"type": "array", "items": {"type": "string"}},
    },
}

##################################
#       classes.json             #
##################################

attribute_groups = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["id", "name", "attributes"],
        "properties": {
            "id": {"type": "integer"},
            "name": {"type": "string"},
            "attributes": {
                "type": "array",
                "itmes": {
                    "type": "object",
                    "required": ["id", "name"],
                    "properties": {"id": {"type": "integer"}, "name": {"type": "string"}},
                },
            },
        },
    },
}

classes_export = {
    "type": "array",
    "items": {
        "type": "object",
        "required": ["name", "id", "attribute_groups"],
        "properties": {"name": {"type": "string"}, "id": {"type": "integer"}, "attribute_groups": attribute_groups},
    },
}



from typing import Any, Dict, List

import deprecation

import darwin.datatypes as dt
from darwin.version import __version__

DEPRECATION_MESSAGE = """

This function is going to be turned into private. This means that breaking 
changes in its interface and implementation are to be expected. We encourage using ``build_image_annotation`` 
instead of calling this low-level function directly.

"""


def build_image_annotation(annotation_file: dt.AnnotationFile) -> Dict[str, Any]:
    """
    Builds and returns a dictionary with the annotations present in the given file.

    Parameters
    ----------
    annotation_file: dt.AnnotationFile
        File with the image annotations to extract.

    Returns
    -------
    Dict[str, Any]
        A dictionary with the annotation from the given file. Has the following structure:

        .. code-block:: python

            {
                "annotations": [
                    {
                        "annotation_type": { ... }, # annotation_data
                        "name": "annotation class name",
                        "bounding_box": { ... } # Optional parameter, only present if the file has a bounding box as well
                    }
                ],
                "image": {
                    "filename": "a_file_name.json",
                    "height": 1000,
                    "width": 2000,
                    "url": "https://www.darwin.v7labs.com/..."
                }
            }
    """
    annotations: List[Dict[str, Any]] = []
    for annotation in annotation_file.annotations:
        payload = {
            annotation.annotation_class.annotation_type: _build_annotation_data(annotation),
            "name": annotation.annotation_class.name,
        }

        if (
            annotation.annotation_class.annotation_type == "complex_polygon"
            or annotation.annotation_class.annotation_type == "polygon"
        ) and "bounding_box" in annotation.data:
            payload["bounding_box"] = annotation.data["bounding_box"]

        annotations.append(payload)

    return {
        "annotations": annotations,
        "image": {
            "filename": annotation_file.filename,
            "height": annotation_file.image_height,
            "width": annotation_file.image_width,
            "url": annotation_file.image_url,
        },
    }


@deprecation.deprecated(
    deprecated_in="0.7.8",
    removed_in="0.8.0",
    current_version=__version__,
    details=DEPRECATION_MESSAGE,
)
def build_annotation_data(annotation: dt.Annotation) -> Dict[str, Any]:
    if annotation.annotation_class.annotation_type == "complex_polygon":
        return {"path": annotation.data["paths"]}

    if annotation.annotation_class.annotation_type == "polygon":
        return dict(filter(lambda item: item[0] != "bounding_box", annotation.data.items()))

    return dict(annotation.data)


def _build_annotation_data(annotation: dt.Annotation) -> Dict[str, Any]:
    if annotation.annotation_class.annotation_type == "complex_polygon":
        return {"path": annotation.data["paths"]}

    if annotation.annotation_class.annotation_type == "polygon":
        return dict(filter(lambda item: item[0] != "bounding_box", annotation.data.items()))

    return dict(annotation.data)
